# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, Iterable, BinaryIO, List, Dict, FrozenSet
import os
import pandas as pd

import numpy as np
import arviz as az
import jax
import jax.numpy as jnp
import pickle

from twinify.mst import MST_selection
from twinify.base import InferenceModel, InferenceResult, InvalidFileFormatException
from twinify.napsu_mq.markov_network import MarkovNetwork
from twinify.napsu_mq.marginal_query import FullMarginalQuerySet
from twinify.dataframe_data import DataFrameData, DataDescription
import twinify.napsu_mq.privacy_accounting as privacy_accounting
import twinify.napsu_mq.maximum_entropy_inference as mei
from twinify.napsu_mq.private_pgm.domain import Domain
from twinify.napsu_mq.private_pgm.dataset import Dataset
import d3p.random


class NapsuMQModel(InferenceModel):
    """Implementation for NAPSU-MQ algorithm, differentially private synthetic data generation method for discrete sensitive data.

    reference: arXiv:2205.14485
    "Noise-Aware Statistical Inference with Differentially Private Synthetic Data", Ossi Räisä, Joonas Jälkö, Samuel Kaski & Antti Honkela
    """

    def __init__(self,
            column_feature_set: Iterable[FrozenSet[str]], # TODO: what are these actually
            use_laplace_approximation: bool = True) -> None:
        """
        Args:
            column_feature_set: ???
            use_laplace_approximation (bool): Use Laplace approximation in the model.
        """

        super().__init__()
        if column_feature_set is None:
            raise ValueError("column_feature_set may not be None")
        # if column_feature_set is None:
        #     column_feature_set = []
        self._column_feature_set = column_feature_set
        self._use_laplace_approximation = use_laplace_approximation

    @property
    def uses_laplace_approximation(self) -> bool:
        return self._use_laplace_approximation

    def fit(self, data: pd.DataFrame, rng: d3p.random.PRNGState, epsilon: float, delta: float,
            query_sets: Optional[Iterable] = None, **kwargs) -> 'NapsuMQResult':
        """Fit differentially private NAPSU-MQ model from data.

        Args:
            data (pd.DataFrame): Pandas Dataframe containing discrete categorical data
            rng (d3p.random.PRNGState): d3p PRNG key
            epsilon (float): Epsilon for differential privacy mechanism
            delta (float): Delta for differential privacy mechanism

        Returns:
            NapsuMQResult: Class containing learned probabilistic model with posterior values
        """
        column_feature_set = self._column_feature_set
        use_laplace_approximation = self._use_laplace_approximation

        dataframe = DataFrameData(data)
        n, d = dataframe.int_df.shape

        if query_sets is None:
            domain_key_list = list(dataframe.columns)
            domain_value_count_list = [len(dataframe.values_by_col[key]) for key in domain_key_list]
            domain = Domain(domain_key_list, domain_value_count_list)
            query_sets = MST_selection(Dataset(dataframe.int_df, domain), epsilon, delta,
                                       cliques_to_include=column_feature_set)

        queries = FullMarginalQuerySet(query_sets, dataframe.values_by_col)
        queries = queries.get_canonical_queries()
        mnjax = MarkovNetwork(dataframe.values_by_col, queries)
        suff_stat = np.sum(queries.flatten()(dataframe.int_df.to_numpy()), axis=0)

        # determine Gaussian mech DP noise level for given privacy level
        sensitivity = np.sqrt(2 * len(query_sets))
        sigma_DP = privacy_accounting.sigma(epsilon, delta, sensitivity)

        # add DP noise (according to Gaussian mechanism)
        inference_rng, dp_rng = d3p.random.split(rng, 2)
        dp_noise = d3p.random.normal(dp_rng, suff_stat.shape) * sigma_DP
        dp_suff_stat = suff_stat + dp_noise

        inference_rng = d3p.random.convert_to_jax_rng_key(inference_rng)

        if use_laplace_approximation is True:
            approx_rng, mcmc_rng = jax.random.split(inference_rng, 2)
            laplace_approx, success = mei.run_numpyro_laplace_approximation(approx_rng, dp_suff_stat, n, sigma_DP,
                                                                            mnjax)
            mcmc, backtransform = mei.run_numpyro_mcmc_normalised(
                mcmc_rng, dp_suff_stat, n, sigma_DP, mnjax, laplace_approx, num_samples=2000, num_warmup=800,
                num_chains=4
            )
            inf_data = az.from_numpyro(mcmc, log_likelihood=False)
            posterior_values = inf_data.posterior.stack(draws=("chain", "draw"))
            posterior_values = backtransform(posterior_values.norm_lambdas.values.transpose())

        else:
            mcmc = mei.run_numpyro_mcmc(
                inference_rng, dp_suff_stat, n, sigma_DP, mnjax, num_samples=2000, num_warmup=800, num_chains=4
            )
            inf_data = az.from_numpyro(mcmc, log_likelihood=False)
            posterior_values = inf_data.posterior.stack(draws=("chain", "draw"))
            posterior_values = posterior_values.lambdas.values.transpose()

        return NapsuMQResult(dataframe.values_by_col, queries, posterior_values, dataframe.data_description)


class NapsuMQResult(InferenceResult):
    """
    NAPSU-MQ result class containing learned differentially private probabilistic model from data.
    Contains functions to generate differentially private synthetic datasets from the original dataset.
    """

    def __init__(self, dataframe_domain: Dict[str, List[int]], queries: 'FullMarginalQuerySet', posterior_values: jnp.ndarray,
                 data_description: DataDescription) -> None:
        super().__init__()
        self._dataframe_domain = dataframe_domain
        self._queries = queries
        self._posterior_values = np.asarray(posterior_values)
        self._data_description = data_description

    @property
    def dataframe_domain(self) -> Dict[str, List[int]]:
        return self._dataframe_domain

    @property
    def queries(self) -> 'FullMarginalQuerySet':
        return self._queries

    @property
    def posterior_values(self) -> np.ndarray:
        return self._posterior_values

    @property
    def data_description(self) -> DataDescription:
        return self._data_description

    def _store_to_io(self, write_io: BinaryIO) -> None:
        assert write_io.writable()
        result = pickle.dumps(self)
        return NapsuMQResultIO.store_to_io(write_io, result)

    @classmethod
    def _load_from_io(cls, read_io: BinaryIO, **kwargs) -> 'NapsuMQResult':
        return NapsuMQResultIO.load_from_io(read_io)

    def generate(
            self,
            rng: d3p.random.PRNGState,
            num_parameter_samples: int,
            num_data_per_parameter_sample: int = 1,
            single_dataframe: bool = True) -> Union[Iterable[pd.DataFrame], pd.DataFrame]:
        jax_rng = d3p.random.convert_to_jax_rng_key(rng)
        mnjax = MarkovNetwork(self._dataframe_domain, self._queries)
        posterior_values = jnp.array(self.posterior_values)
        inds = jax.random.choice(key=jax_rng, a=posterior_values.shape[0], shape=[num_parameter_samples])
        posterior_sample = posterior_values[inds, :]
        rng, *data_keys = jax.random.split(jax_rng, num_parameter_samples + 1)
        syn_datasets = [mnjax.sample(syn_data_key, jnp.array(posterior_value), num_data_per_parameter_sample) for
                        syn_data_key, posterior_value
                        in list(zip(data_keys, posterior_sample))]

        dataframes = [self.data_description.map_to_categorical(syn_data) for syn_data in syn_datasets]

        if single_dataframe is True:
            combined_dataframe = pd.concat(dataframes, ignore_index=True)
            return combined_dataframe
        else:
            return dataframes


class NapsuMQResultIO:
    IDENTIFIER = "NapsuMQ".encode("utf8")
    CURRENT_IO_VERSION = 1
    # Replace with twinify.serialization.ENDIANESS with merge
    CURRENT_IO_VERSION_BYTES = CURRENT_IO_VERSION.to_bytes(1, 'big')

    @staticmethod
    def load_from_io(read_io: BinaryIO) -> NapsuMQResult:
        assert read_io.readable()

        if not NapsuMQResultIO.is_file_stored_result_from_io(read_io, reset_cursor=False):
            raise InvalidFileFormatException(NapsuMQResult, "Stored data does not have correct type identifier.")

        current_version = int.from_bytes(read_io.read(1), 'big')
        if current_version != NapsuMQResultIO.CURRENT_IO_VERSION:
            raise InvalidFileFormatException(NapsuMQResult, "Stored data uses an unknown storage format version.")

        result_binary = read_io.read()
        result = pickle.loads(result_binary)

        return result

    @staticmethod
    def is_file_stored_result_from_io(read_io: BinaryIO, reset_cursor: bool) -> bool:
        assert read_io.readable()
        assert read_io.seekable()

        identifier = read_io.read(len(NapsuMQResultIO.IDENTIFIER))
        if reset_cursor:
            read_io.seek(-len(identifier), os.SEEK_CUR)

        if identifier == NapsuMQResultIO.IDENTIFIER:
            return True

        return False

    @staticmethod
    def store_to_io(write_io: BinaryIO, result) -> None:
        assert write_io.writable()

        write_io.write(NapsuMQResultIO.IDENTIFIER)
        write_io.write(NapsuMQResultIO.CURRENT_IO_VERSION_BYTES)
        write_io.write(result)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Iterable

import graphviz


class UndirectedGraph:
    """An undirected graph datastructure."""

    def __init__(self, nodes: Iterable, edges: Iterable):
        """Create the graph by specifying nodes and edges explicitly.
        Args:
            nodes (iterable): The nodes.
            edges (iterable): The edges as a list of pairs of nodes.
        """
        self.nodes = set(nodes)
        self.edges = set(edges)
        self.check_nodes_edges()

    def check_nodes_edges(self):
        for e1, e2 in self.edges:
            edge = (e1, e2)
            if e1 not in self.nodes:
                raise ValueError("{} from {} not in nodes".format(e1, edge))
            if e2 not in self.nodes:
                raise ValueError("{} from {} not in nodes".format(e2, edge))
            if (e2, e1) in self.edges:
                raise ValueError("Duplicate edge {}".format(edge))

    def visualize(self):
        """Visualize the graph with Graphviz.
        Returns:
            Graphviz object: The graph as a Graphviz object.
        """
        g = graphviz.Graph()
        for e1, e2 in self.edges:
            g.edge(str(e1), str(e2))
        for node in self.nodes:
            g.node(str(node))
        return g

    def from_edge_list(edges):
        """Create an undirected graph from a list of edges.
        Args:
            edges (list): List of tuples forming the edges.
        Returns:
            UndirectedGraph: The resulting UndirectedGraph.
        """
        nodes = set()
        for e1, e2 in edges:
            nodes.add(e1)
            nodes.add(e2)
        return UndirectedGraph(nodes, edges)

    def from_clique_list(cliques):
        """Create an undirected graph from a list of cliques.
        Args:
            cliques (list): List of cliques as a list of tuples of nodes.
        Returns:
            UndirectedGraph: The resulting UndirectedGraph.
        """
        nodes = set()
        edges = set()
        for clique in cliques:
            for i, node in enumerate(clique):
                nodes.add(node)
                for j in range(i + 1, len(clique)):
                    edges.add((node, clique[j]))

        return UndirectedGraph(nodes, edges)

    def copy(self):
        nodes = self.nodes.copy()
        edges = self.edges.copy()
        return UndirectedGraph(nodes, edges)

    def add_edge(self, edge):
        e1, e2 = edge
        if (e2, e1) not in self.edges:
            self.edges.add(edge)
            self.add_node(e1)
            self.add_node(e2)

    def add_node(self, node):
        self.nodes.add(node)

    def get_neighbours(self, node):
        return [
            other
            for other in self.nodes
            if (node, other) in self.edges or (other, node) in self.edges
        ]


def node_ordering_cost(graph: UndirectedGraph, node: Iterable):
    return len(graph.get_neighbours(node))


def greedy_ordering(graph: UndirectedGraph) -> List:
    """Find a greedy ordering for variable elimination.
    Args:
        graph (UndirectedGraph): The graph to find the ordering on.
    Returns:
        list: The elimination order as a list of nodes.
    """
    graph = graph.copy()
    unmarked_nodes = set(graph.nodes)
    ordering = []
    for i in range(len(unmarked_nodes)):
        min_cost = 2**32
        min_cost_node = None
        for node in unmarked_nodes:
            cost = node_ordering_cost(graph, node)
            if cost < min_cost:
                min_cost = cost
                min_cost_node = node

        ordering.append(min_cost_node)
        unmarked_nodes.remove(min_cost_node)

        neighbours = graph.get_neighbours(min_cost_node)
        for i, neighbour in enumerate(neighbours):
            for other_neighbour in neighbours[i + 1 :]:
                graph.add_edge((neighbour, other_neighbour))

    return ordering



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna, © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/557c077708d3559212a8f65dff3eccd3fd244abb/src/mbi/clique_vector.py
# Modified by the twinify Developers under the Apache 2.0 license
# Changed import statements

import numpy as np
from twinify.napsu_mq.private_pgm.factor import Factor


class CliqueVector(dict):
    """ This is a convenience class for simplifying arithmetic over the
        concatenated vector of marginals and potentials.
        These vectors are represented as a dictionary mapping cliques (subsets of attributes)
        to marginals/potentials (Factor objects)
    """

    def __init__(self, dictionary):
        self.dictionary = dictionary
        dict.__init__(self, dictionary)

    @staticmethod
    def zeros(domain, cliques):
        return CliqueVector({cl: Factor.zeros(domain.project(cl)) for cl in cliques})

    @staticmethod
    def ones(domain, cliques):
        return CliqueVector({cl: Factor.ones(domain.project(cl)) for cl in cliques})

    @staticmethod
    def uniform(domain, cliques):
        return CliqueVector({cl: Factor.uniform(domain.project(cl)) for cl in cliques})

    @staticmethod
    def random(domain, cliques, prng=np.random):
        return CliqueVector({cl: Factor.random(domain.project(cl), prng) for cl in cliques})

    @staticmethod
    def normal(domain, cliques, prng=np.random):
        return CliqueVector({cl: Factor.normal(domain.project(cl), prng) for cl in cliques})

    @staticmethod
    def from_data(data, cliques):
        ans = {}
        for cl in cliques:
            mu = data.project(cl)
            ans[cl] = Factor(mu.domain, mu.datavector())
        return CliqueVector(ans)

    def combine(self, other):
        # combines this CliqueVector with other, even if they do not share the same set of factors
        # used for warm-starting optimization
        # Important note: if other contains factors not defined within this CliqueVector, they
        # are ignored and *not* combined into this CliqueVector
        for cl in other:
            for cl2 in self:
                if set(cl) <= set(cl2):
                    self[cl2] += other[cl]
                    break

    def __mul__(self, const):
        ans = {cl: const * self[cl] for cl in self}
        return CliqueVector(ans)

    def __rmul__(self, const):
        return self.__mul__(const)

    def __add__(self, other):
        if np.isscalar(other):
            ans = {cl: self[cl] + other for cl in self}
        else:
            ans = {cl: self[cl] + other[cl] for cl in self}
        return CliqueVector(ans)

    def __sub__(self, other):
        return self + -1 * other

    def exp(self):
        ans = {cl: self[cl].exp() for cl in self}
        return CliqueVector(ans)

    def log(self):
        ans = {cl: self[cl].log() for cl in self}
        return CliqueVector(ans)

    def dot(self, other):
        return sum((self[cl] * other[cl]).sum() for cl in self)

    def size(self):
        return sum(self[cl].domain.size() for cl in self)




# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mixture model used by twinify main script and available for modelling.
"""

import numpy as np
import jax.numpy as jnp
import jax

import numpyro
import numpyro.distributions as dists

from jax.scipy.special import logsumexp
from numpyro.distributions.constraints import Constraint
from numpyro.distributions.transforms import biject_to
import typing


class _CombinedConstraint(Constraint):
    """
    Wraps a constraint by aggregating over ``reinterpreted_batch_ndims``-many
    dims in :meth:`check`, so that an event is valid only if all its
    independent entries are valid.
    """

    def __init__(self,
            base_constraints: typing.Iterable[Constraint],
            sizes: typing.Optional[typing.Iterable[int]] = None) -> None:

        self._base_constraints = list()

        event_dims = 1
        for i, base_constraint in enumerate(base_constraints):
            if not isinstance(base_constraint, Constraint):
                raise ValueError(f"Element {i} in constraint list is not of type Constraint.")

            if base_constraint.event_dim > 1:
                raise ValueError(
                    f"Currently only support base constraints with event_dim <= 1 but element {i}"
                    f"in constraint list has event_dim = {base_constraint.event_dim}"
                )

            self._base_constraints.append(base_constraint)

        self._event_dims = event_dims
        if sizes is not None:
            sizes = np.array(sizes)
            if len(sizes) != len(self._base_constraints):
                raise ValueError(
                    f"Number of sizes {len(sizes)} has to equal number of base constraints {(len(self._base_constraints))}."
                )

            self._dim_offsets = tuple(np.cumsum(np.hstack((0, sizes))))
        else:
            self._dim_offsets = tuple(range(len(self._base_constraints) + 1))

        assert len(self._dim_offsets) == len(self._base_constraints) + 1

        super().__init__()

    def __call__(self, value: jnp.ndarray) -> jnp.ndarray:
        if jnp.shape(value)[-1] != self.size:
            raise ValueError(f"Last dimension of value ({jnp.shape(value)[-1]})"\
                f"does match the constraint size ({self.size})!")

        base_results = [
            jnp.all(constraint(value[..., start:end]))
            for constraint, start, end in zip(self._base_constraints, self._dim_offsets[:-1], self._dim_offsets[1:])
        ]
        return jnp.all(jnp.stack(base_results, -1))

    def feasible_like(self, prototype: jnp.ndarray) -> jnp.ndarray:
        if jnp.shape(prototype)[-1] != self.size:
            raise ValueError(f"Last dimension of value ({jnp.shape(prototype)[-1]})"\
                f"does match the constraint size ({self.size})!")

        base_results = [
            constraint.feasible_like(prototype[..., start:end]).reshape(jnp.shape(prototype)[:-1] + (end - start,))
            for constraint, start, end in zip(self._base_constraints, self._dim_offsets[:-1], self._dim_offsets[1:])
        ]

        results = jnp.concatenate(base_results, axis=-1)
        return results

    @property
    def is_discrete(self) -> bool:
        return np.all(tuple(constraint.is_discrete for constraint in self._base_constraints))

    @property
    def event_dim(self) -> int:
        return self._event_dims

    @property
    def size(self) -> int:
        return self._dim_offsets[-1]

    @property
    def offsets(self) -> typing.Iterable[int]:
        return self._dim_offsets[:-1]


combined_constraint = _CombinedConstraint
# TODO: need to implement a transform for this

class MixtureModel(dists.Distribution):
    """ A general purpose mixture model.

    Described a number of distributions that independently model feature dimensions
    of the data.
    The MixtureModel then is a mixture of components which all follow the
    structure described by the joint feature distributions.

    Feature distributions therefore must have a leading batch dimension that corresponds
    to the number of mixture components.

    Note this class behaves quite differently from `numpyro.distributions.Mixture`,
    where each distribution describes all features for a single mixture component.
     """

    arg_constraints = {
        'pis' : dists.constraints.simplex
    }

    has_enumerate_support = False
    reparametrized_params = ["pis"]

    @property
    def support(self) -> Constraint:
        return self._constraint

    @property
    def pis(self) -> jnp.ndarray:
        return self._pis

    def __init__(self, distributions: typing.Iterable[dists.Distribution], pis: np.ndarray, *, validate_args=None):
        """
        Initializes a MixtureModel instance.

        The feature distributions provided by argument `dists` must be in the
        same order as the features are provided in the `value` vector passed
        to `log_prob`.

        Args:
            distributions (list of numpyro.distributions.Distribution): The list of feature distributions.
            pis (array_like): The mixture weights.
        """
        self.distributions = list(distributions)
        batch_shape = None
        constraints = []
        sizes = []

        self._pis = pis
        self._mixture_components = len(pis)

        for dist in distributions:
            if not isinstance(dist, dists.Distribution):
                raise ValueError("MixtureModel got an argument that is not a distribution.")
            if len(dist.event_shape) > 1:
                raise ValueError(f"Feature distribution event shape cannot have more than one dimensions, was {dist.event_shape}.")

            if batch_shape is None:
                batch_shape = dist.batch_shape
                if len(batch_shape) == 0 or batch_shape[-1] != self._mixture_components:
                    raise ValueError(
                        f"Last batch dimension must equal the number of mixture components ({self._mixture_components})"
                        f" for all distributions; got {dist.batch_shape} from {dist}.")
            else:
                if batch_shape != dist.batch_shape:
                    raise ValueError(f"Feature distribution batch shape deviates from mixture batch shape {batch_shape}; got {dist.batch_shape} from {dist}.")

            sizes.append(dist.event_shape[-1] if len(dist.event_shape) > 0 else 1)
            constraints.append(dist.support)

        assert len(batch_shape) > 0
        batch_shape = batch_shape[:-1]
        self._offsets = np.cumsum(np.hstack((0., sizes)), dtype=np.int32)
        event_shape = (self._offsets[-1],)

        if len(pis) != self._mixture_components:
            raise ValueError(f"Mixture model has {self._mixture_components} components but only got {len(pis)} mixture weights.")

        self._constraint = combined_constraint(constraints, sizes)

        super(MixtureModel, self).__init__(batch_shape, event_shape, validate_args=validate_args)

    @property
    def mixture_components(self):
        return self._mixture_components

    def log_prob(self, value):
        """
        Evaluates the log-probabilities for given observations.

        The log-probabilities of the K component mixture model
        $\log p(x | \theta)  =  \log \sum_{k=1}^K \pi_k p(x | \theta_k)$
        where $p(x | \theta_k)$ denotes the probabilities of the k-th mixture cluster

        Args:
            values (array_like): The observations. Dimension of the last axis must
                be the number of features (i.e., `len(dists)`).
        Returns:
            array_like of shape `values.shape[:-1]`
        """
        log_phis = [
            dbn.log_prob(value[..., np.newaxis, start:stop])
            if len(dbn.event_shape) > 0
            else dbn.log_prob(value[..., np.newaxis, start])
            for dbn, start, stop in zip(self.distributions, self._offsets[:-1], self._offsets[1:])
        ]
        log_phis = jnp.array(log_phis).sum(0)
        log_pis = jnp.log(self._pis)

        temp = log_pis + log_phis

        return logsumexp(temp, axis=-1)

    def sample(self, key, sample_shape=()):
        """
        Samples from the mixture model.

        Args:
            key (jax.random.PRNGKey): RNG key.
            sample_shape (tuple): How many samples to draw and how to arrange them
        Returns:
            array_like of shape `sample_shape + self.batch_shape + self.event_shape`
        """
        return self.sample_with_intermediates(key, sample_shape)[0]

    def sample_with_intermediates(self, key, sample_shape=()):
        batch_shape = sample_shape + self.batch_shape

        vals_rng_key, pis_rng_key = jax.random.split(key, 2)
        zs = dists.Categorical(self._pis).sample(pis_rng_key, sample_shape=batch_shape).flatten()

        component_keys = jax.random.split(vals_rng_key, len(self.distributions))
        vals = []
        for dbn, dbn_key in zip(self.distributions, component_keys):
            dbn_sample = dbn.sample(dbn_key, sample_shape=sample_shape)
            dbn_event_shape = dbn.event_shape
            if dbn_event_shape == ():
                dbn_event_shape = (1,)

            flat_batch_size = int(np.prod(batch_shape))

            dbn_sample = dbn_sample.reshape(flat_batch_size, self._mixture_components, *dbn_event_shape)
            dbn_sample = dbn_sample[np.arange(flat_batch_size), zs]
            dbn_sample = dbn_sample.reshape(*batch_shape, *dbn_event_shape)
            vals.append(dbn_sample)

        return jnp.concatenate(vals, axis=-1), [zs.reshape(batch_shape)]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import numpy as np
from jax.numpy import ndarray
from numpy.typing import ArrayLike
from typing import List, Optional, Tuple, Union, Sequence, Iterable


class LogFactor:
    """Jax implementation of log factors of Markov networks."""

    def __init__(self, scope: Iterable[int], values: jnp.ndarray, debug_checks: Optional[bool] = True):
        """Create the log factor.
        Args:
            scope (List): Scope of the factor as a tuple of nodes.
            values (array): Values of the factor as an array where values[i, j, k] is the value for output values (i, j, k).
            debug_checks (bool, optional): If enabled, check that some operations are valid for the factor. Defaults to True.
        """
        self.scope: Tuple[int] = tuple(scope)
        self.variable_inds = {var: i for i, var in enumerate(scope)}
        self.values = values
        self.debug_checks = debug_checks

    def get_variable_range(self, variable: Union[str, int]) -> Tuple[int, ...]:
        return self.values.shape[self.variable_inds[variable]]

    def repeat_to_shape(self, result_shape: Union[List[int], List[Tuple[int, ...]]], result_scope: Tuple[int]):
        variables_not_in_scope: List[int] = [var for var in result_scope if var not in self.scope]
        repeat_shape = [result_shape[result_scope.index(var)] for var in variables_not_in_scope] + [1 for var in
                                                                                                    self.scope]
        repeat_variable_order = variables_not_in_scope + list(self.scope)
        permutation = [repeat_variable_order.index(var) for var in result_scope]
        repeated_values = self.tile_values(repeat_shape)
        return self.permute_axes(repeated_values, permutation)

    def product(self, factor: 'LogFactor') -> 'LogFactor':
        """Product of self with another factor.
        Args:
            factor (LogFactor): The other factor.
        Raises:
            ValueError: If debug checks are enabled, check that the ranges for the variables of both factors are equal.
        Returns:
            LogFactor: The resulting product as a new LogFactor object.
        """
        if self.debug_checks:
            common_variables = set(self.scope).intersection(set(factor.scope))
            for common_var in common_variables:
                if self.get_variable_range(common_var) != factor.get_variable_range(common_var):
                    raise ValueError("Ranges for variable {} are not equal. {} != {}".format(
                        common_var, self.get_variable_range(common_var), factor.get_variable_range(common_var)
                    ))

        if set(self.scope) == set(factor.scope):
            permutation = [factor.scope.index(var) for var in self.scope]
            result_scope = self.scope
            result_values = self.values + factor.permute_axes(factor.values, permutation)
        else:
            result_scope = tuple(set(self.scope).union(set(factor.scope)))
            result_values_ranges = [
                self.get_variable_range(var) if var in self.scope else factor.get_variable_range(var) for var in
                result_scope
            ]
            self_expanded_values = self.repeat_to_shape(result_values_ranges, result_scope)
            other_expanded_values = factor.repeat_to_shape(result_values_ranges, result_scope)
            result_values = self_expanded_values + other_expanded_values

        return LogFactor(result_scope, result_values, self.debug_checks)

    def marginalise(self, variable: int) -> 'LogFactor':
        """Marginalise a variable from the factor.
        Args:
            variable (int): The variable to marginalise.
        Returns:
            LogFactor: The marginalised factor as a new LogFactor object.
        """
        to_remove_index = self.scope.index(variable)
        result_scope = self.scope[:to_remove_index] + self.scope[to_remove_index + 1:]
        result_values = self.logsumexp(self.values, axis=self.variable_inds[variable])
        return LogFactor(result_scope, result_values, self.debug_checks)

    def log_sum_total(self) -> ndarray:
        """Compute the logsumexp of all values in self.
        Returns:
            float: The resulting logsumexp value.
        """
        return self.logsumexp(self.values, tuple(range(len(self.values.shape))))

    def list_product(factors: Iterable['LogFactor']) -> 'LogFactor':
        """The product of a list of LogFactors.
        Args:
            factors (list(LogFactor)): The list of factors in the product.
        Returns:
            LogFactor: The product of the factors as a new LogFactor object.
        """
        factors = list(factors)
        product = factors[0]
        for i in range(1, len(factors)):
            product = product.product(factors[i])
        return product

    def project(self, variables: List) -> 'LogFactor':
        """Marginalise all variables except the given variables.
        Args:
            variables (list): The variables that are not marginalised.
        Raises:
            ValueError: If debug checks are enabled, checks that variables is a subset of self.scope.
        Returns:
            LogFactor: The result as a new LogFactor object.
        """
        if self.debug_checks and not set(variables).issubset(set(self.scope)):
            raise ValueError("Variables {} are not a subset of scope {}".format(variables, self.scope))

        to_eliminate = set(self.scope).difference(set(variables + ["batch"]))
        factor = self
        for var in to_eliminate:
            factor = factor.marginalise(var)
        return factor

    def condition(self, variable: int, value: int) -> 'LogFactor':
        """Condition the factor on a given value for a given variable.
        Args:
            variable (int): The variable to condition.
            value (int): The value to condition on.
        Returns:
            LogFactor: The conditioned factor as a new LogFactor object.
        """
        to_remove_index = self.scope.index(variable)
        result_scope = self.scope[:to_remove_index] + self.scope[to_remove_index + 1:]
        result_values = self.take_values(value, to_remove_index)
        return LogFactor(result_scope, result_values, self.debug_checks)

    def add_batch_dim(self, n_batches: int) -> 'LogFactor':
        """Add a dimension named 'batch' the scope for batched operations.
        self.values is copied so that each batch has the same values.
        Args:
            n_batches (int): The number of batches.
        Raises:
            ValueError: If the 'batch' variable is already in scope.
        Returns:
            LogFactor: The batched factor as a new LogFactor object.
        """
        batch_dim_name = "batch"
        if batch_dim_name in self.scope:
            raise ValueError("{} is already in scope".format(batch_dim_name))
        result_scope = (batch_dim_name,) + self.scope
        result_values = self.tile_values((n_batches,) + tuple(1 for _ in range(len(self.scope))))
        return LogFactor(result_scope, result_values, self.debug_checks)

    def batch_condition(self, variable: int, values: jnp.ndarray) -> 'LogFactor':
        """Condition a batched LogFactor on an array of values for a given variable.
        Args:
            variable (int): The variable to condition on.
            values (array): The values for each batch to condition on.
        Returns:
            LogFactor: The result as a new LogFactor object.
        """
        self.ensure_batch_is_first_dim()
        to_remove_index = self.scope.index(variable)
        result_scope = self.scope[:to_remove_index] + self.scope[to_remove_index + 1:]
        result_shape = [self.get_variable_range(var) for var in result_scope]
        result_values = self.compute_batch_condition_values(values, to_remove_index, result_shape)
        return LogFactor(result_scope, result_values, self.debug_checks)

    def ensure_batch_is_first_dim(self) -> None:
        if self.scope[0] != "batch" and "batch" in self.scope:
            batch_index = self.variable_inds["batch"]
            self.scope = ("batch",) + self.scope[:batch_index] + self.scope[batch_index + 1:]
            self.move_values_axis(batch_index, 0)
            self.variable_inds = {var: i for i, var in enumerate(self.scope)}

    def tile_values(self, repeat_shape: ArrayLike) -> jnp.ndarray:
        return jnp.tile(self.values, repeat_shape)

    def permute_axes(self, value: jnp.ndarray, permutation: Union[None, Tuple[int, ...]]) -> jnp.ndarray:
        return value.transpose(permutation)

    def logsumexp(self, values: ArrayLike, axis: Union[None, Tuple[int, ...]]) -> jnp.ndarray:
        return jax.scipy.special.logsumexp(values, axis=axis)

    def take_values(self, index: Union[Iterable, np.number], axis: Optional[Union[int, np.number]]) -> jnp.ndarray:
        return jnp.take(self.values, index, axis)

    def compute_batch_condition_values(self, values: ArrayLike, to_remove_index: int,
                                       result_shape=None) -> jnp.ndarray:
        return jax.vmap(lambda i: jnp.take(self.values[i], values[i], to_remove_index - 1), 0, 0)(
            jnp.arange(self.get_variable_range("batch")))

    def move_values_axis(self, axis: Union[int, Sequence[int]], place: Union[int, Sequence[int]]) -> None:
        self.values = jnp.moveaxis(self.values, axis, place)

    def query(self, queries) -> Union[jnp.ndarray, jnp.number]:
        result = jnp.zeros(len(queries.queries))
        for i, query in enumerate(queries.queries):
            query_permutation = [query.features.index(variable) for variable in self.scope]
            query_value_tuple = query.value_tuple()
            result = result.at[i].set(self.values[tuple(query_value_tuple[i] for i in query_permutation)])
        return jnp.exp(result - self.log_sum_total())



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional, Tuple, Dict, Set

import itertools
import numpy as np

from tqdm import tqdm

from twinify.napsu_mq.utils import powerset


class MarginalQuery:
    def __init__(self, inds: Iterable[int], value: Iterable[int], features: Optional[Iterable] = None):
        """Create the marginal query object.
        Args:
            inds (iterable(int)): Indices of the marginal query. Converted to list.
            value (iterable(int)): Values of the marginal query. Converted to np.ndarray.
            features (list, optional): Variable names corresponding to the indices. Defaults to inds.
        """
        self.inds = list(inds)
        self.features = self.inds if features is None else features
        self.value = value if type(value) is np.ndarray else np.array(value)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return (x[:, self.inds] == self.value).all(axis=1).astype(int).reshape((-1, 1))

    def __str__(self) -> str:
        return "{} = {}".format(self.inds, [val.item() for val in self.value])

    def as_tuple(self) -> Tuple:
        """Convert to tuple for easy equality comparisons.
        Returns:
            tuple: (indices, value)
        """
        return tuple(self.inds), self.value_tuple()

    def value_tuple(self) -> Tuple:
        """Convert self.value to tuple.
        Returns:
            tuple: self.value as a tuple.
        """
        return tuple([val.item() for val in self.value])


class FullMarginalQuerySet:
    """A full marginal query set."""

    def __init__(self, feature_sets: Iterable[Tuple], values_by_feature: Dict):
        """Create the full marginal query set.
        Args:
            feature_sets (list(tuple)): The tuples of variables that are used as indices.
            values_by_feature (dict): A dict containing a list of possible values for each variable.
        """
        self.feature_sets = list(feature_sets)
        self.values_by_feature = values_by_feature

        self.feature_by_index = list(self.values_by_feature.keys())

        self.int_feature_sets = [
            tuple(self.feature_by_index.index(feature) for feature in feature_set)
            for feature_set in self.feature_sets
        ]

        self.values_by_int_feature = {
            self.feature_by_index.index(feature): values
            for feature, values in self.values_by_feature.items()
        }

        self.queries = {feature_set: QueryList(all_marginals_for_feature_set(feature_set, values_by_feature)) for
                        feature_set in feature_sets}

    def query(self, x: np.ndarray) -> Dict:
        """Run all marginal queries on output.
        Args:
            x (np.ndarray): The output.
        Returns:
            dict: A dictionary containing the feature sets as keys and query results as values.
        """
        return {feature_set: self.queries[feature_set](x) for feature_set in self.feature_sets}

    def query_sum(self, x: np.ndarray) -> Dict:
        """As self.query, except the results are summed over datapoints.
        Args:
            x (np.ndarray): The output.
        Returns:
            dict: A dictionary containing the feature sets as keys and summed query results as values.
        """
        return {feature_set: result.sum(axis=0) for feature_set, result in self.query(x).items()}

    def query_feature_set(self, feature_set: Tuple, x: np.ndarray) -> np.ndarray:
        """Query one feature set.
        Args:
            feature_set (tuple): The feature set to query.
            x (np.ndarray): The output.
        Returns:
            np.ndarray: The query results.
        """
        return self.queries[feature_set](x)

    def query_feature_set_sum(self, feature_set: Tuple, x: np.ndarray) -> np.ndarray:
        """As self.query_feature_set, except results are summed over datapoints.
        Args:
            feature_set (tuple): The feature set to query.
            x (np.ndarray): The output.
        Returns:
            np.ndarray: The summed query results.
        """
        return self.query_feature_set(feature_set, x).sum(axis=0)

    def flatten(self) -> 'QueryList':
        """Convert self to a QueryList of the contained queries.
        Returns:
            QueryList: The resulting QueryList.
        """
        return QueryList(itertools.chain.from_iterable(query_list.queries for query_list in self.queries.values()))

    def get_canonical_queries(self) -> 'FullMarginalQuerySet':
        """Find the canonical queries for the queries in self.
        Returns:
            FullMarginalQuerySet: The canonical queries as a FullMarginalQuerySet.
        """
        d = len(self.values_by_feature.keys())
        base_value = np.zeros(d, dtype=int)
        clique_list = list(itertools.chain.from_iterable(powerset(features) for features in self.int_feature_sets))
        clique_set = set([tuple(val) for val in clique_list])
        canonical_queries = {}

        for clique in tqdm(clique_set):
            clique = set(clique)
            clique_ordered = tuple(clique)
            if clique == set():
                continue
            index_conversion = np.full((d,), -1)
            for i, variable in enumerate(clique_ordered):
                index_conversion[variable] = i
            conv_clique_indices = index_conversion[list(clique)]

            clique_product = itertools.product(*(self.values_by_int_feature[variable] for variable in clique_ordered))
            for val in tqdm(clique_product):
                value = np.zeros(len(clique), dtype=int)
                value[conv_clique_indices] = np.asarray(val, dtype=int)
                counter = np.zeros(tuple([len(self.values_by_int_feature[variable]) for variable in clique_ordered]),
                                      dtype=int)
                subsets = powerset(clique)
                for subset in subsets:
                    multiplier = (-1) ** (len(clique) - len(subset))
                    base_indices = list(clique.difference(subset))
                    conv_base_indices = index_conversion[base_indices]
                    completed_value = value.copy()
                    completed_value[conv_base_indices] = base_value[conv_base_indices]
                    counter[tuple(completed_value)] += multiplier

                clique_ordered_not_int = tuple(self.feature_by_index[feature] for feature in clique_ordered)
                if not (counter == 0).all():
                    if clique_ordered_not_int not in canonical_queries.keys():
                        canonical_queries[clique_ordered_not_int] = []
                    canonical_queries[clique_ordered_not_int].append(
                        MarginalQuery(clique_ordered, val, features=clique_ordered_not_int)
                    )

        added_query_tuples = {query.as_tuple() for query in itertools.chain.from_iterable(canonical_queries.values())}
        not_original_clique_queries = list(itertools.chain.from_iterable(
            queries for clique, queries in canonical_queries.items() if clique not in self.feature_sets))
        original_clique_queries = {clique: queries for clique, queries in canonical_queries.items() if
                                   clique in self.feature_sets}

        for clique in tqdm(self.feature_sets):
            if clique not in original_clique_queries.keys():
                original_clique_queries[clique] = []

        for query in tqdm(not_original_clique_queries):
            original_clique = None
            for clique in self.feature_sets:
                if set(query.features).issubset(set(clique)):
                    original_clique = clique
                    break
            features_to_sum = set(clique).difference(set(query.features))
            values_to_sum = itertools.product(*(self.values_by_feature[feature] if feature in features_to_sum else (
                query.value[query.features.index(feature)],) for feature in original_clique))
            new_queries = [
                MarginalQuery(tuple(self.feature_by_index.index(feature) for feature in original_clique), value,
                              features=original_clique)
                for value in values_to_sum
            ]
            for new_query in new_queries:
                if new_query.as_tuple() not in added_query_tuples:
                    added_query_tuples.add(new_query.as_tuple())
                    original_clique_queries[original_clique].append(new_query)

        canonical_queries = {key: QueryList(queries) for key, queries in original_clique_queries.items()}
        new_fmqs = FullMarginalQuerySet([], self.values_by_feature)
        new_fmqs.queries = canonical_queries
        new_fmqs.feature_sets = list(canonical_queries.keys())
        return new_fmqs


class QueryList:
    """A list of MarginalQuery objects."""

    def __init__(self, queries: Iterable[MarginalQuery]):
        """Create the QueryList.
        Args:
            queries (iterable(MarginalQuery)): The MarginalQuery objects.
        """
        self.queries = list(queries)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.concatenate([query(x) for query in self.queries], axis=1)

    def __str__(self):
        return "\n".join(query.__str__() for query in self.queries)

    def evaluate_subset(self, x: np.ndarray, subset: Iterable[int]) -> np.ndarray:
        """Evaluate a subset of the queries.
        Args:
            x (np.ndarray): The output.
            subset (iterable(int)): The indices for the queries to evaluate.
        Returns:
            np.ndarray: The query results.
        """
        return np.concatenate([self.queries[i](x) for i in subset], axis=1)

    def get_subset(self, subset: Iterable[int]) -> 'QueryList':
        """Get a subset of the queries.
        Args:
            subset (iterable(int)): The indices for the queries to return.
        Returns:
            QueryList: The selected queries as a new QueryList object.
        """
        return QueryList([self.queries[i] for i in subset])

    def as_tuple_set(self) -> Set[Tuple]:
        """Get a set of tuples representing the queries.
        Returns:
            set(tuple): A set of tuples of each query.
        """
        return {query.as_tuple() for query in self.queries}

    def get_variable_associations(self, use_features: Optional[bool] = False) -> Dict:
        """Get a dict containing a list of indices into self.queries for each set of indices that self.queries look at.
        Args:
            use_features (bool, optional): Use query features as keys instead of query indices. Defaults to False.
        Returns:
            dict: A dict with tuples of query indices or features as keys and a list of indices of queries with that set of indices or features.
        """
        result = {}
        for i, query in enumerate(self.queries):
            inds_tuple = tuple(query.inds if not use_features else query.features)
            if inds_tuple in result.keys():
                result[inds_tuple].append(i)
            else:
                result[inds_tuple] = [i]
        return result


def join_query_sets(query_sets: Iterable[QueryList]) -> QueryList:
    """Combine several query lists.
    Args:
        query_sets (iterable(QueryList)): The QueryLists to combine.
    Returns:
        QueryList: The combined QueryList.
    """
    return QueryList(list(itertools.chain.from_iterable(query_set.queries for query_set in query_sets)))


def all_marginals_for_feature_set(feature_set: Iterable, values_by_feature: Dict) -> List[MarginalQuery]:
    inds = [list(values_by_feature.keys()).index(feature) for feature in feature_set]
    all_values = list(itertools.product(*[sorted(values_by_feature[feature]) for feature in feature_set]))
    return [MarginalQuery(inds, value, feature_set) for value in all_values]


def all_marginals(feature_sets: Iterable[Tuple], values_by_feature: Dict) -> QueryList:
    """Get all marginal queries for given sets of variables.
    Args:
        feature_sets (list(tuple)): The sets of variables.
        values_by_feature (dict): A dict containing the possible values for each variable.
    Returns:
        QueryList: All marginals for the given sets of variables as a QueryList.
    """
    queries = []
    for feature_set in feature_sets:
        queries_for_feature_set = all_marginals_for_feature_set(feature_set, values_by_feature)
        for query in queries_for_feature_set:
            queries.append(query)
    return QueryList(queries)


def column_feature_sets_to_indices(column_feature_sets: Iterable[Tuple[str, str]],
                                   columns: List[str]) -> List[Tuple[int, int]]:
    int_feature_sets = []
    for pair in column_feature_sets:
        if pair[0] not in columns:
            raise ValueError(f"Variable {pair[0]} not found in the column list")
        if pair[1] not in columns:
            raise ValueError(f"Variable {pair[1]} not found in the column list")

        int_feature_sets.append((
            columns.index(pair[0]), columns.index(pair[1])
        ))

    return int_feature_sets



#!/usr/bin/env python

# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Twinify main script.
"""

from jax.config import config
config.update("jax_enable_x64", True)

from twinify.cli.dpvi_numpyro_model_loading import ModelException

import numpy as np
import pandas as pd
from typing import Optional, Tuple

import argparse
import d3p.random

from twinify.base import InferenceModel, InferenceResult
from twinify.dpvi import InferenceException
from twinify.cli.dpvi_loader import load_cli_dpvi
from twinify.cli.napsu_loader import load_cli_napsu
from twinify import DataDescription

from twinify import __version__

def add_common_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument('data_path', type=str, help='Path to input data.')
    parser.add_argument('model_path', type=str, help='Path to model file (.txt or .py).')
    parser.add_argument("output_path", type=str, help="Path prefix to outputs (synthetic data and model).")
    parser.add_argument("--epsilon", default=1., type=float, help="Target multiplicative privacy parameter epsilon.")
    parser.add_argument("--delta", default=None, type=float, help="Target additive privacy parameter delta.")
    parser.add_argument("--seed", default=None, type=int, help="PRNG seed used in model fitting. If not set, will be securely initialized to a random value.")
    parser.add_argument("--num_synthetic", "--n", default=None, type=int, help="Amount of synthetic data to generate in total. By default as many as input data.")
    parser.add_argument("--num_synthetic_records_per_parameter_sample", "--m", default=1, type=int, help="Amount of synthetic samples to sample per parameter value drawn from the learned parameter posterior.")
    parser.add_argument("--drop_na", action='store_true', default=False, help="Remove missing values from data.")
    parser.add_argument("--separate_output", default=False, action='store_true', help="Store synthetic data in separate files per parameter sample.")

parser = argparse.ArgumentParser(description='Twinify: Program for creating synthetic twins under differential privacy.',\
        fromfile_prefix_chars="%")
subparsers = parser.add_subparsers(title='inference_method', dest='method')
parser.add_argument("--version", action='version', version=__version__)


## ARGUMENT PARSING FOR DPVI
vi_parser = subparsers.add_parser("vi")
add_common_arguments(vi_parser)
vi_parser.add_argument("--clipping_threshold", default=1., type=float, help="Clipping threshold for DP-SGD.")
vi_parser.add_argument("--k", default=50, type=int, help="Mixture components in fit (for automatic modelling only).")
vi_parser.add_argument("--num_epochs", "-e", default=200, type=int, help="Number of training epochs.")
vi_parser.add_argument("--sampling_ratio", "-q", default=0.01, type=float, help="Subsampling ratio for DP-SGD.")
vi_parser.set_defaults(load_fn=load_cli_dpvi)

## ARGUMENT PARSING FOR NAPSUMQ
napsu_parser = subparsers.add_parser("napsumq")
add_common_arguments(napsu_parser)
napsu_parser.set_defaults(load_fn=load_cli_napsu)


def initialize_rngs(seed: Optional[int] = None) -> Tuple[d3p.random.PRNGState, d3p.random.PRNGState]:
    master_rng = d3p.random.PRNGKey(seed)
    print(f"RNG seed: {seed}")

    inference_rng, sampling_rng, numpyro_seed = d3p.random.split(master_rng, 3)

    numpyro_seed = int(d3p.random.random_bits(numpyro_seed, 32, (1,)))
    np.random.seed(numpyro_seed)

    return inference_rng, sampling_rng

def main():
    args, unknown_args = parser.parse_known_args()
    if unknown_args:
        print(f"Additional received arguments: {unknown_args}")

    # read data
    try:
        train_df = pd.read_csv(args.data_path)
    except Exception as e:
        print("#### UNABLE TO READ DATA FILE ####")
        print(e)
        return 1
    print("Loaded data set has {} rows (entries) and {} columns (features).".format(*train_df.shape))
    num_data = len(train_df)

    data_description = DataDescription.from_dataframe(train_df)

    try:
        model: InferenceModel = args.load_fn(args, unknown_args, data_description)

        # check DP values
        target_delta = args.delta
        if target_delta is None:
            target_delta = 0.1 / num_data
        if target_delta * num_data > 1.:
            print("!!!!! WARNING !!!!! The given value for privacy parameter delta ({:1.3e}) exceeds 1/(number of data) ({:1.3e}),\n" \
                "which is the maximum value that is usually considered safe!".format(
                    target_delta, 1. / num_data
                ))
            x = input("Continue? (type YES ): ")
            if x != "YES":
                print("Aborting...")
                return 4
            print("Continuing... (YOU HAVE BEEN WARNED!)")

        inference_rng, sampling_rng = initialize_rngs(args.seed)

        try:
            dpvi_result: InferenceResult = model.fit(
                train_df,
                inference_rng,
                args.epsilon,
                target_delta,
                verbose=True
            )
        except (InferenceException, FloatingPointError):
            print("################################## ERROR ##################################")
            print("!!!!! The inference procedure encountered a NaN value (not a number). !!!!!")
            print("This means the model has major difficulties in capturing the data and is")
            print("likely to happen when the dataset is very small and/or sparse.")
            print("Try adapting (simplifying) the model.")
            print("Aborting...")
            return 2

        # Store learned model parameters
        # TODO: we should have a mode for twinify that allows to rerun the sampling without training, using stored parameters
        dpvi_result.store(f"{args.output_path}.p")
        # TODO: previous twinify would store the entire provided arguments, for reproducability; this is not the case anymore; can we somehow reinstate this?

        # sample synthetic data
        print("Model learning complete; now sampling data!")
        num_synthetic = args.num_synthetic
        if num_synthetic is None:
            num_synthetic = num_data

        num_parameter_samples = int(np.ceil(num_synthetic / args.num_synthetic_records_per_parameter_sample))
        num_synthetic = num_parameter_samples * args.num_synthetic_records_per_parameter_sample
        print(f"Will sample {args.num_synthetic_records_per_parameter_sample} synthetic data records for each of "
              f"{num_parameter_samples} samples from the parameter posterior for a total of {num_synthetic} records.")
        if args.separate_output:
            print("They will be stored as separate data sets for each parameter posterior sample.")
        else:
            print("They will be stored as a single large data set.")

        syn_data = dpvi_result.generate(
            sampling_rng,
            num_parameter_samples,
            args.num_synthetic_records_per_parameter_sample,
            single_dataframe=not args.separate_output
        )

        # store synthetic data
        if isinstance(syn_data, pd.DataFrame):
            syn_data.to_csv(args.output_path + ".csv", index=False)
        else:
            for i, syn_df in enumerate(syn_data):
                syn_df.to_csv(f"{args.output_path}.{i}.csv", index=False)

        ### illustrate results TODO need to adopt new way of handing train_df
        #if args.visualize != 'none':
        #    show_popups = args.visualize in ('popup', 'both')
        #    save_plots = args.visualize in ('store', 'both')
        #    # Missing value rate
        #    if not args.drop_na:
        #        missing_value_fig = plot_missing_values(syn_df, train_df, show=show_popups)
        #        if save_plots:
        #            missing_value_fig.savefig(args.output_path + "_missing_value_plots.svg")
        #    # Marginal violins
        #    margin_fig = plot_margins(syn_df, train_df, show=show_popups)
        #    # Covariance matrices
        #    cov_fig = plot_covariance_heatmap(syn_df, train_df, show=show_popups)
        #    if save_plots:
        #        margin_fig.savefig(args.output_path + "_marginal_plots.svg")
        #        cov_fig.savefig(args.output_path + "_correlation_plots.svg")
        #    if show_popups:
        #        plt.show()
        return 0
    except ModelException as e:
        print(e.format_message(args.model_path))
    except AssertionError as e:
        raise e
    except Exception as e:
        print("#### AN UNCATEGORISED ERROR OCCURRED ####")
        raise e
    return 1

if __name__ == "__main__":
    exit(main())



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle

from typing import BinaryIO, Optional, Dict, Union, Iterable, Tuple, Any

import pandas as pd

import numpy as np
from numpy.typing import ArrayLike
import jax
import jax.numpy as jnp
import numpyro

import d3p.random
from twinify.base import InferenceResult, InvalidFileFormatException
from twinify.dpvi import ModelFunction, GuideFunction, PrivacyLevel
import twinify.serialization
from twinify.dpvi.sampling import sample_synthetic_data
from twinify.dataframe_data import DataDescription


class SamplingException(Exception):
    pass


class DPVIResult(InferenceResult):

    def __init__(self,
            model: ModelFunction,
            guide: GuideFunction,
            parameters: Dict[str, ArrayLike],
            privacy_parameters: PrivacyLevel,
            final_elbo: float,
            data_description: DataDescription,
        ) -> None:

        self._model = model
        self._guide = guide
        self._params = parameters
        self._privacy_level = privacy_parameters
        self._final_elbo = final_elbo
        self._data_description = data_description

    _twinify_model_output_site = '_twinify_output'

    @staticmethod
    def _mark_model_outputs(model: ModelFunction) -> ModelFunction:
        def _model_wrapper(data: Optional[ArrayLike]=None, *args, **kwargs) -> Any:
            """ Wraps a model function and captures the return value in a sampling site named `_twinify_output`,
            which `DPVIResult` uses to read the generated synthetic data from.
            """
            samples = model(*args, **kwargs)
            if len(jnp.shape(samples)) != 2:
                raise SamplingException("A numpyro model for twinify must return the sampled data as a single two-dimensional array.")
            numpyro.deterministic(DPVIResult._twinify_model_output_site, samples)
            return samples
        return _model_wrapper

    def generate(self,
            rng: d3p.random.PRNGState,
            num_parameter_samples: int,
            num_data_per_parameter_sample: int = 1,
            single_dataframe: bool = True
        ) -> Union[Iterable[pd.DataFrame], pd.DataFrame]:

        sampling_rng = d3p.random.convert_to_jax_rng_key(rng)
        samples = sample_synthetic_data(
            self._mark_model_outputs(self._model), self._guide,
            self._params, sampling_rng,
            num_parameter_samples,
            num_data_per_parameter_sample
        )

        samples = samples[self._twinify_model_output_site]

        assert samples.shape[:2] == (num_parameter_samples, num_data_per_parameter_sample)

        def _squash_sample_dims(v: np.array) -> np.array:
            old_shape = np.shape(v)
            new_shape = (old_shape[0] * old_shape[1], *old_shape[2:])
            reshaped_v = np.reshape(v, new_shape)
            return reshaped_v

        if single_dataframe:
            samples_df = self._data_description.map_to_categorical(_squash_sample_dims(samples))
            return samples_df
        else:
            return [
                self._data_description.map_to_categorical(samples[i]) for i in range(num_parameter_samples)
            ]

    @classmethod
    def _load_from_io(
            cls, read_io: BinaryIO,
            model: ModelFunction,
            guide: Optional[GuideFunction] = None,
            **kwargs
        ) -> 'InferenceResult':

        if guide is None:
            from twinify.dpvi.dpvi_model import DPVIModel
            guide = DPVIModel.create_default_guide(model)

        parameters, privacy_parameters, final_elbo, data_description = DPVIResultIO.load_params_from_io(read_io)

        return DPVIResult(model, guide, parameters, privacy_parameters, final_elbo, data_description)

    @classmethod
    def _is_file_stored_result_from_io(cls, read_io: BinaryIO) -> bool:
        return DPVIResultIO.is_file_stored_result_from_io(read_io, reset_cursor=True)

    def _store_to_io(self, write_io: BinaryIO) -> None:
        return DPVIResultIO.store_params_to_io(
            write_io, self._params, self._privacy_level, self._final_elbo, self._data_description
        )

    @property
    def guide(self) -> GuideFunction:
        return self._guide

    @property
    def model(self) -> ModelFunction:
        return self._model

    @property
    def parameters(self) -> Dict[str, ArrayLike]:
        return jax.tree_map(lambda x: np.copy(x), self._params)

    @property
    def privacy_level(self) -> float:
        """ The privacy parameters: epsilon, delta and standard deviation of noise applied during inference. """
        return self._privacy_level

    @property
    def final_elbo(self) -> float:
        """ The final ELBO achieved by the inference (on the training data). """
        return self._final_elbo

    @property
    def data_description(self) -> DataDescription:
        return self._data_description


class DPVIResultIO:

    IDENTIFIER = "DPVI".encode("utf8")
    CURRENT_IO_VERSION = 1
    CURRENT_IO_VERSION_BYTES = CURRENT_IO_VERSION.to_bytes(1, twinify.serialization.ENDIANESS)

    @staticmethod
    # def load_from_io(read_io: BinaryIO, treedef: jax.tree_util.PyTreeDef) -> DPVIResult:
    def load_params_from_io(read_io: BinaryIO) -> Tuple[Dict[str, ArrayLike], Iterable[str]]:
        assert read_io.readable()

        if not DPVIResultIO.is_file_stored_result_from_io(read_io, reset_cursor=False):
            raise InvalidFileFormatException(DPVIResult, "Stored data does not have correct type identifier.")

        current_version = int.from_bytes(read_io.read(1), twinify.serialization.ENDIANESS)
        if current_version != DPVIResultIO.CURRENT_IO_VERSION:
            raise InvalidFileFormatException(DPVIResult, "Stored data uses an unknown storage format version.")

        # parameters = twinify.serialization.read_params(read_io, treedef)
        stored_data = pickle.load(read_io)
        return (
            stored_data['params'],
            stored_data['privacy_level'],
            stored_data['final_elbo'],
            stored_data['data_description'],
        )

    @staticmethod
    def is_file_stored_result_from_io(read_io: BinaryIO, reset_cursor: bool) -> bool:
        assert read_io.readable()
        assert read_io.seekable()

        identifier = read_io.read(len(DPVIResultIO.IDENTIFIER))
        if reset_cursor:
            read_io.seek(-len(identifier), os.SEEK_CUR)

        if identifier == DPVIResultIO.IDENTIFIER:
            return True

        return False

    @staticmethod
    def store_params_to_io(
            write_io: BinaryIO,
            params: Dict[str, ArrayLike],
            privacy_level: PrivacyLevel,
            final_elbo: float,
            data_description: DataDescription
        ) -> None:

        assert write_io.writable()

        data = {
            'params': params,
            'privacy_level': privacy_level,
            'final_elbo': final_elbo,
            'data_description': data_description,
        }

        write_io.write(DPVIResultIO.IDENTIFIER)
        write_io.write(DPVIResultIO.CURRENT_IO_VERSION_BYTES)
        pickle.dump(data, write_io)
        # twinify.serialization.write_params(params, write_io)



#!/usr/bin/env python

# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Twinify tools main script.
"""

import argparse
import twinify.cli.tools.check_model
from twinify import __version__

parser = argparse.ArgumentParser(description='Twinify tools: Various twinify utility scripts.')
parser.add_argument("--version", action='version', version=__version__)
subparsers = parser.add_subparsers(dest='command')
twinify.cli.tools.check_model.setup_argument_parser(subparsers.add_parser('check-model', help='Checks whether a given model works with twinify.'))

def main():
    args, unknown_args = parser.parse_known_args()

    if args.command =='check-model':
        return twinify.cli.tools.check_model.main(args, unknown_args)
    else:
        print('#### UNKNOWN OPERATING MODE ####')
        print(f'## twinify-tools does not know the command {args.command}')
        return 1

if __name__ == "__main__":
    exit(main())



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings as _warnings
from jax.config import config as _config
_config.update("jax_enable_x64", True)
import jax.numpy as _jnp
if _jnp.array([1.]).dtype != _jnp.float64:
    _warnings.warn(
        "Failed to set floating point precision to 64 bits. You may experience instabilities "
        "with twinify as a result.\n"
        "This can happen when the jax library is already initialized in your scripts before twinify is loaded. "
        "In that case, please enable 64 bit precision by including the following before the first usage of jax:\n"
        "   from jax.config import config\n"
        "   config.update('jax_enable_x64', True).",
        stacklevel=2
    )

import twinify.dpvi
import twinify.napsu_mq
from twinify.dataframe_data import DataDescription
from twinify.version import VERSION
__version__ = VERSION
from twinify.base import InferenceModel, InferenceResult



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
import inspect
from re import A
import traceback
import pandas as pd
import numpy as np
from typing import Any, Callable, Tuple, Iterable, Dict, Union, Optional, Sequence
import twinify.dpvi
from twinify.dpvi.modelling import automodel
from numpyro.infer.autoguide import AutoDiagonalNormal
from twinify import DataDescription
import os
import argparse

__all__ = ['load_custom_numpyro_model', 'ModelException', 'NumpyroModelParsingUnknownException', 'NumpyroModelParsingException']

TPreprocessFunction = Callable[[pd.DataFrame], Union[pd.DataFrame, pd.Series]]
TGuardedPreprocessFunction = Callable[[pd.DataFrame], pd.DataFrame]
TPostprocessFunction = Callable[[pd.DataFrame], Union[pd.DataFrame, pd.Series]]
TGuardedPostprocessFunction = Callable[[pd.DataFrame], pd.DataFrame]
TModelFunction = twinify.dpvi.ModelFunction
TGuideFunction = twinify.dpvi.GuideFunction
TModelFactoryFunction = Callable[[argparse.Namespace, Iterable[str], DataDescription], Union[TModelFunction, Tuple[TModelFunction, TGuideFunction]]]


class ModelException(Exception):

    @staticmethod
    def filter_traceback(tb: Iterable[traceback.StackSummary], model_file_path: str) -> Iterable[traceback.StackSummary]:
        assert model_file_path is not None
        model_file_name = os.path.basename(model_file_path)
        encountered_model = False
        for frame in tb:
            if frame.filename.endswith(model_file_name):
                encountered_model = True
            if encountered_model:
                yield frame

    def __init__(self, title: str, msg: str=None, base_exception: Optional[Exception]=None) -> None:
        self.msg = msg if msg is not None else "Uncategorised error"
        self.base = base_exception
        self.title = title

        super().__init__(self.format_message())

    def format_message(self, model_file_path: Optional[str]=None) -> str:
        full_message = f"\n#### {self.title.upper()} ####\n##   {self.msg}"
        if self.base is not None:
            full_message += f"\nTechnical error description below:\n"
            tb = traceback.extract_tb(self.base.__traceback__)
            if model_file_path is not None:
                tb = self.filter_traceback(tb, model_file_path)
            full_message += "\n".join(traceback.format_list(tb))
            full_message += "\n".join(traceback.format_exception_only(type(self.base), self.base))

        return full_message

class NumpyroModelParsingException(ModelException):

    def __init__(self, msg: str, base_exception: Optional[Exception]=None) -> None:
        self.msg = msg
        self.base = base_exception

        title = "FAILED TO PARSE THE MODEL SPECIFICATION"

        super().__init__(title, msg, base_exception)

class NumpyroModelParsingUnknownException(NumpyroModelParsingException):

    def __init__(self, function_name: str, base_exception: Exception) -> None:
        self.__init__(f"Uncategorised error while trying to access function '{function_name}' from model module.", base_exception)


def guard_preprocess(preprocess_fn: TPreprocessFunction) -> TGuardedPreprocessFunction:
    def wrapped_preprocess(train_df: pd.DataFrame) -> pd.DataFrame:
        try:
            train_data = preprocess_fn(train_df)
        except TypeError as e:
            if str(e).find('positional argument') != -1:
                raise ModelException("FAILED DURING PREPROCESSING DATA", "Custom preprocessing functions must accept a single pandas.DataFrame as argument.")
            else:
                raise ModelException("FAILED DURING PREPROCESSING DATA", base_exception=e) from e
        except Exception as e:
            raise ModelException("FAILED DURING PREPROCESSING DATA", base_exception=e) from e

        if not isinstance(train_data, (pd.DataFrame, pd.Series)):
            raise ModelException("FAILED DURING PREPROCESSING DATA", f"Custom preprocessing functions must return a pandas.DataFrame or pandas.Series, but returned a {type(train_data)} instead.")

        if isinstance(train_data, pd.Series):
            train_data = train_data.to_frame()

        return train_data

    return wrapped_preprocess

@guard_preprocess
def default_preprocess(train_df: pd.DataFrame) -> Tuple[pd.DataFrame, int]:
    return train_df

def guard_postprocess(postprocess_fn: TPostprocessFunction) -> TGuardedPostprocessFunction:
    def wrapped_postprocess(posterior_samples: pd.DataFrame) -> pd.DataFrame:
        try:
            retval = postprocess_fn(posterior_samples)
        except TypeError as e:
            if str(e).find('positional argument') != -1:
                raise ModelException("FAILED DURING POSTPROCESSING DATA", "Custom postprocessing functions must accept a pandas.DataFrame as argument.")
            else:
                raise e
        except Exception as e:
            raise ModelException("FAILED DURING POSTPROCESSING DATA", base_exception=e) from e

        if not isinstance(retval, (pd.DataFrame, pd.Series)):
            raise ModelException("FAILED DURING POSTPROCESSING DATA", f"Custom postprocessing functions must return a single pandas.DataFrame (or Series); got {type(retval)}.")

        if isinstance(retval, pd.Series):
            retval = retval.to_frame()
        return retval
    return wrapped_postprocess

def guard_model(model_fn: TModelFunction) -> TModelFunction:
    def wrapped_model(*args, **kwargs) -> Any:
        try:
            return model_fn(*args, **kwargs)
        except TypeError as e:
            if str(e).find('positional argument') != -1 or str(e).find('num_obs_total') != -1:
                raise ModelException("FAILED IN MODEL", f"Custom model functions can accept any number of positional arguments for training but must default all of them to None for synthesising data. They also must accept the num_obs_total keyword argument.", base_exception=e) from e
            else:
                raise ModelException("FAILED IN MODEL", base_exception=e) from e
        except Exception as e:
            raise ModelException("FAILED IN MODEL", base_exception=e) from e
    return wrapped_model

def load_custom_numpyro_model(
        model_path: str, args: argparse.Namespace, unknown_args: Iterable[str], orig_data: pd.DataFrame
    ) -> Tuple[TModelFunction, TGuideFunction, TGuardedPreprocessFunction, TGuardedPostprocessFunction]:

    if not os.path.exists(model_path):
        raise FileNotFoundError(model_path)

    try:
        spec = importlib.util.spec_from_file_location("model_module", model_path)
        model_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(model_module)
    except Exception as e: # handling errors in py-file parsing
        raise NumpyroModelParsingException("Unable to read the specified file as a Python module.", e) from e

    # load the model function from the module
    model = None
    guide = None
    try: model = model_module.model
    except AttributeError:
        # model file did not directly contain a model function; check if it has model_factory
        try:
            model_factory = model_module.model_factory
        except AttributeError:
            raise NumpyroModelParsingException("Model module does neither specify a 'model' nor a 'model_factory' function.")
        try:
            model_factory_return = model_factory(args, unknown_args, orig_data)
        except TypeError as e:
            if str(e).find('positional argument') != -1:
                raise ModelException("FAILED IN MODEL FACTORY", f"Custom model_factory functions must accept a namespace of parsed arguments, an iterable of unparsed arguments and a pandas.DataFrame as arguments.")
            raise e
        except Exception as e:
            raise ModelException('FAILED IN MODEL FACTORY', base_exception=e) from e

        # determine whether model_factory returned model function or (model, guide) tuple
        if (type(model_factory_return) is tuple
                and isinstance(model_factory_return[0], TModelFunction)
                and isinstance(model_factory_return[1], TGuideFunction)):
            model, guide = model_factory_return
        elif isinstance(model_factory_return, TModelFunction):
            model = model_factory_return
        else:
            raise ModelException('FAILED IN MODEL FACTORY', f"Custom model_factory functions must return either a model function or a tuple consisting of model and guide function, but returned {type(model_factory_return)}.")
    except Exception as e: raise NumpyroModelParsingUnknownException('model', e) from e

    if not isinstance(model, Callable):
        raise NumpyroModelParsingException(f"'model' must be a function; got {type(model)}")
    model = guard_model(model)

    if guide is None:
        try: guide = model_module.guide
        except AttributeError: guide = AutoDiagonalNormal(model)
        except Exception as e: raise NumpyroModelParsingUnknownException('guide', e) from e

    # try to obtain preprocessing function from custom model
    try: preprocess_fn = guard_preprocess(model_module.preprocess)
    except AttributeError: preprocess_fn = default_preprocess
    except Exception as e: raise NumpyroModelParsingUnknownException('preprocess', e) from e

    # try to obtain postprocessing function from custom model
    try: postprocess_fn = guard_postprocess(model_module.postprocess)
    except AttributeError:
        print("Warning: Your model does not specify a postprocessing function for generated samples.")
        print("     Using default, which assumes that your model only produces samples at sample site 'x' and outputs samples as they are.")
        postprocess_fn = automodel.postprocess_function_factory([])
    except Exception as e: raise NumpyroModelParsingUnknownException('postprocess', e) from e

    return model, guide, preprocess_fn, postprocess_fn



#!/usr/bin/env python

# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Twinify custom model checking mode script.
"""

import argparse
from typing import Iterable

import jax

import pandas as pd
import numpy as np
import numpyro
from numpyro.handlers import trace, seed
from numpyro.infer import Predictive
from twinify.cli.dpvi_loader import load_cli_dpvi
from twinify import DataDescription
from twinify.base import InferenceModel
from twinify.cli.preprocessing_model import PreprocessingModel
from twinify.dpvi.dpvi_model import DPVIModel, DPVIResult
from twinify.cli.dpvi_numpyro_model_loading import ModelException
import d3p.random

from twinify.dpvi.sampling import sample_synthetic_data

def setup_argument_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument('data_path', type=str, help='Path to input data.')
    parser.add_argument('model_path', type=str, help='Path to model file (.py).')
    parser.add_argument("--drop_na", default=False, action='store_true', help="Remove missing values from data.")
    parser.add_argument("--full_traceback", default=False, action='store_true', help="Print a full traceback when errors occur, instead of filtering for custom model code.")

    # we mirror all arguments of the twinify main script here as models may now use any of these
    # in the model_factory method. They are not used by the check-model script.
    parser.add_argument("--epsilon", default=1., type=float, help="[UNUSED] Target multiplicative privacy parameter epsilon.")
    parser.add_argument("--delta", default=None, type=float, help="[UNUSED] Target additive privacy parameter delta.")
    parser.add_argument("--clipping_threshold", default=1., type=float, help="[UNUSED] Clipping threshold for DP-SGD.")
    parser.add_argument("--seed", default=None, type=int, help="[UNUSED] PRNG seed used in model fitting. If not set, will be securely initialized to a random value.")
    parser.add_argument("--k", default=50, type=int, help="[UNUSED] Mixture components in fit (for automatic modelling only).")
    parser.add_argument("--num_epochs", "-e", default=2, type=int, help="[UNUSED] Number of training epochs.")
    parser.add_argument("--sampling_ratio", "-q", default=0.01, type=float, help="[UNUSED] Subsampling ratio for DP-SGD.")
    parser.add_argument("--num_synthetic", "--n", default=None, type=int, help="Amount of synthetic data to generate in total. By default as many as input data.")
    parser.add_argument("--num_synthetic_records_per_parameter_sample", "--m", default=1, type=int, help="Amount of synthetic samples to sample per parameter value drawn from the learned parameter posterior.")



def main(args: argparse.Namespace, unknown_args: Iterable[str]) -> int:
    # read data
    try:
        train_df = pd.read_csv(args.data_path)
    except Exception as e:
        print("#### UNABLE TO READ DATA FILE ####")
        print(e)
        exit(1)

    args.output_path = ''
    args.num_epochs = 2

    data_description = DataDescription.from_dataframe(train_df)

    num_data = 100
    train_df = train_df.iloc[:num_data]


    try:
        model: PreprocessingModel = load_cli_dpvi(args, unknown_args, data_description)

        preprocessed_train_df = model.preprocess(train_df)
        data_description = DataDescription.from_dataframe(preprocessed_train_df)

        dpvi_model: DPVIModel = model._base_model

        print("Sampling from prior distribution (using model, guide)")
        # We use Preditive with model to sample from the prior predictive distribution. Since this does not inolve guide,
        # Predictive has no clue about which of the samples are for observations and which are for parameter values.
        # Since we expect postprocess_fn to deal only with observations, we trace through guide to identify
        # parameter sample sites and filter those out. (To invoke guide we need a small batch of data, for which we
        # use whatever preprocess_fn returned to get the right shapes, but zero it out to prevent information leakage).
        try:
            prior_samples = Predictive(DPVIResult._mark_model_outputs(dpvi_model._model), num_samples = num_data)(jax.random.PRNGKey(0))
        except Exception as e: raise ModelException("Error while obtaining prior samples from model", base_exception=e)

        prior_samples = prior_samples[DPVIResult._twinify_model_output_site].squeeze(1)

        print("Transforming prior samples to output domain to obtain dummy data (using postprocess)")
        # prior_samples = pd.DataFrame({site: np.asarray(samples.squeeze(1)) for site, samples in prior_samples.items() if site not in parameter_sites})
        prior_samples = pd.DataFrame(prior_samples, columns=data_description.columns)
        prior_samples = data_description.map_to_categorical(prior_samples)
        if model.postprocess_fn is not None:
            prior_samples = model.postprocess_fn(prior_samples)

        assert len(prior_samples) == num_data # TODO: maybe not?

        print("Preprocessing dummy data (using preprocess) and inferring model parameters (using model, guide)")
        try:
            result = model.fit(
                prior_samples,
                d3p.random.PRNGKey(13),
                epsilon=10.,
                delta=0.1
            )
        except Exception as e:
            raise ModelException("Error while performing inference", base_exception=e)

        print("Sampling from posterior distribution (using model, guide) and postprocessing (using postprocess)")
        try:
            posterior_samples = result.generate(d3p.random.PRNGKey(0), num_data, num_data)
        except Exception as e:
            raise ModelException("Error while obtaining posterior samples from model", base_exception=e)

        print("Everything okay!")
        return 0

    except ModelException as e:
        if args.full_traceback:
            print(e)
        else:
            print(e.format_message(args.model_path))
    except AssertionError as e:
        raise e
    except Exception as e:
        print("#### AN UNCATEGORISED ERROR OCCURRED ####")
        raise e
    return 1

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Twinify: Program for creating synthetic twins under differential privacy.',\
        fromfile_prefix_chars="%")
    setup_argument_parser(parser)
    exit(
        main(parser.parse_known_args())
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna, © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/master/mechanisms/mst.py
# Modified by the twinify Developers under the Apache 2.0 license
# Modifications contain adding typing hints and changing privacy accounting method

from typing import Iterable, List, Tuple, Callable, Mapping

import numpy as np
import scipy.sparse
from twinify.napsu_mq.private_pgm.inference import FactoredInference
from twinify.napsu_mq.private_pgm.dataset import Dataset
from twinify.napsu_mq.private_pgm.domain import Domain
from scipy import sparse
from disjoint_set import DisjointSet
import networkx as nx
import itertools
import twinify.napsu_mq.privacy_accounting as accounting
from scipy.special import logsumexp

"""
This is a generalization of the winning mechanism from the
2018 NIST Differential Privacy Synthetic Data Competition.

Unlike the original implementation, this one can work for any discrete dataset,
and does not rely on public provisional data for measurement selection.
"""


def MST_selection(data: Dataset, epsilon: float, delta: float, cliques_to_include: Iterable[Tuple[str, str]] = []) -> List:
    """Select marginal queries from dataset and cliques

    Args:
        data (Dataset): Dataset for selection
        epsilon (float): Epsilon for DP mechanism
        delta (float): Delta for DP mechanism
        cliques_to_include (List[Tuple]): Cliques to include in the queries

    Returns:
        Marginal queries for probabilistic model
    """

    rho = accounting.eps_delta_budget_to_rho_budget(epsilon, delta)
    sigma = np.sqrt(3 / (2 * rho))
    cliques = [(col,) for col in data.domain]
    log1 = measure(data, cliques, sigma)
    data, log1, undo_compress_fn = compress_domain(data, log1)
    cliques = select(data, rho / 3.0, log1, cliques=cliques_to_include)
    return cliques


def MST(data: Dataset, epsilon: float, delta: float) -> Dataset:
    """MST algorithm to generate synthetic data from sensitive dataset

    Args:
        data (Dataset): Dataset for MST
        epsilon (float): Epsilon for DP mechanism
        delta (float): Delta for DP mechanism

    Returns:
        Synthetic dataset
    """
    rho = accounting.eps_delta_budget_to_rho_budget(epsilon, delta)
    sigma = np.sqrt(3 / (2 * rho))
    cliques = [(col,) for col in data.domain]
    log1 = measure(data, cliques, sigma)
    data, log1, undo_compress_fn = compress_domain(data, log1)
    cliques = select(data, rho / 3.0, log1)
    log2 = measure(data, cliques, sigma)
    engine = FactoredInference(data.domain, iters=5000)
    est = engine.estimate(log1 + log2)
    synth = est.synthetic_data()
    return undo_compress_fn(synth)


def measure(data: Dataset, cliques: List, sigma: float, weights: np.ndarray = None) -> List[
    Tuple[scipy.sparse.coo_matrix, np.ndarray, float, List]]:
    """Measure marginals with noisy measurements

    Args:
        data (Dataset): Dataset
        cliques (List): Cliques to measure
        sigma (float): Noise scale
        weights (np.ndarray): Weight for each measurement

    Returns:
        Measurement log of noisy measurements from marginals
    """

    if weights is None:
        weights = np.ones(len(cliques))
    weights = np.array(weights) / np.linalg.norm(weights)
    measurements = []
    for proj, wgt in zip(cliques, weights):
        x = data.project(proj).datavector()
        y = x + np.random.normal(loc=0, scale=sigma / wgt, size=x.size)
        Q = sparse.eye(x.size)
        measurements.append((Q, y, sigma / wgt, proj))
    return measurements


def compress_domain(data: Dataset, measurements: Iterable[Tuple]) -> Tuple[Dataset, List, Callable]:
    """Compress domain for dataset

    Args:
        data (Dataset): Dataset to compress
        measurements (List[Tuple]): Measurement log

    Returns:
        Dataset in new domain, new measurements, function to undo compression
    """

    supports = {}
    new_measurements = []
    for Q, y, sigma, proj in measurements:
        col = proj[0]
        sup = y >= 3 * sigma
        supports[col] = sup
        if supports[col].sum() == y.size:
            new_measurements.append((Q, y, sigma, proj))
        else:  # need to re-express measurement over the new domain
            y2 = np.append(y[sup], y[~sup].sum())
            I2 = np.ones(y2.size)
            I2[-1] = 1.0 / np.sqrt(y.size - y2.size + 1.0)
            y2[-1] /= np.sqrt(y.size - y2.size + 1.0)
            I2 = sparse.diags(I2)
            new_measurements.append((I2, y2, sigma, proj))
    undo_compress_fn = lambda data: reverse_data(data, supports)
    return transform_data(data, supports), new_measurements, undo_compress_fn


def exponential_mechanism(q: np.ndarray, eps: float, sensitivity: float, sampling_func: Callable = np.random.choice,
                          monotonic=False) -> np.ndarray:
    """Exponential mechanism for differential privacy

    Args:
        q (np.ndarray):
        eps (float): Epsilon for DP mechanism
        sensitivity (float): Sensitivity of the function
        sampling_func (function): Function to generate random sample
        monotonic: Is monotonic function

    Returns:
        Selected highly weighted pair with differential privacy
    """

    coef = 1.0 if monotonic else 0.5
    scores = coef * eps / sensitivity * q
    probas = np.exp(scores - logsumexp(scores))
    return sampling_func(q.size, p=probas)


def select(data: Dataset, rho: float, measurement_log: Iterable[Tuple],
           cliques: Iterable[Tuple[str, str]] = []) -> List:
    """Select the low dimensional marginals from dataset and measurements

    Args:
        data (Dataset): Dataset for selection
        rho (float): Privacy parameter for rho-zCDP
        measurement_log (Iterable[Tuple]): Log of noisy measurements
        cliques (Iterable[Tuple]): Cliques to include

    Returns:
        List of marginal queries
    """
    engine = FactoredInference(data.domain, iters=1000)
    est = engine.estimate(measurement_log)

    weights = {}
    candidates = list(itertools.combinations(data.domain.attrs, 2))
    for a, b in candidates:
        xhat = est.project([a, b]).datavector()
        x = data.project([a, b]).datavector()
        weights[a, b] = np.linalg.norm(x - xhat, 1)

    T = nx.Graph()
    T.add_nodes_from(data.domain.attrs)
    ds = DisjointSet()

    for e in cliques:
        T.add_edge(*e)
        ds.union(*e)

    r = len(list(nx.connected_components(T)))
    epsilon = np.sqrt(8 * rho / (r - 1))
    for i in range(r - 1):
        candidates = [e for e in candidates if not ds.connected(*e)]
        wgts = np.array([weights[e] for e in candidates])
        idx = exponential_mechanism(wgts, epsilon, sensitivity=1.0)
        e = candidates[idx]
        T.add_edge(*e)
        ds.union(*e)

    return list(T.edges)


def transform_data(data: Dataset, supports: Mapping) -> Dataset:
    """Transform dataset to new domain

    Args:
        data (Dataset): Dataset
        supports (Support): Support

    Returns:
        Dataset in new domain
    """
    df = data.df.copy()
    newdom = {}
    for col in data.domain:
        support = supports[col]
        size = support.sum()
        newdom[col] = int(size)
        if size < support.size:
            newdom[col] += 1
        mapping = {}
        idx = 0
        for i in range(support.size):
            mapping[i] = size
            if support[i]:
                mapping[i] = idx
                idx += 1
        assert idx == size
        df[col] = df[col].map(mapping)
    newdom = Domain.fromdict(newdom)
    return Dataset(df, newdom)


def reverse_data(data: Dataset, supports: Mapping) -> Dataset:
    """Dataset expressed in new domain

    Args:
        data (Dataset): Dataset
        supports (Mapping): Supports for new domain

    Returns:
        Dataset with new domain
    """
    df = data.df.copy()
    newdom = {}
    for col in data.domain:
        support = supports[col]
        mx = support.sum()
        newdom[col] = int(support.size)
        idx, extra = np.where(support)[0], np.where(~support)[0]
        mask = df[col] == mx
        if extra.size == 0:
            pass
        else:
            df.loc[mask, col] = np.random.choice(extra, mask.sum())
        df.loc[~mask, col] = idx[df.loc[~mask, col]]
    newdom = Domain.fromdict(newdom)
    return Dataset(df, newdom)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper routines for modelling.
"""

from typing import Optional
import jax.numpy as jnp
import jax

def slice_feature(
        x: Optional[jnp.ndarray],
        feature_idx_start: int,
        feature_idx_limit: Optional[int] = None,
        stride: int = 1,
        dtype: Optional[jnp.dtype] = None
    ):
    """
    Returns a slice over the feature dimension (axis 1) of a data set. Returns
    `None` without failure if `None` is input.

    If `feature_idx_limit` is `None` (and `x` is an array), returns all columns starting
    from  `feature_idx_start`, i.e., x[:, feature_idx_start::stride]`.

    If `feature_idx_limit` is an integer, returns the slice indicated by the arguments, i.e.,
    `x[:, feature_idx_start:feature_idx_limit:stride]`.

    Args:
        x (jax.numpy.ndarray, None): Array or None.
        feature_idx_start (int) : Feature index of where to start the slice.
        feature_idx_limit (int, None): Optional feature index where to stop the slice.
        stride (int): Optional stride, defaults to 1.
        dtype (jnp.dtype, None): Optional output dtype to convert the slice to.
    """
    if x is None:
        return None

    if dtype is None:
        dtype = jnp.dtype(x)

    return jax.lax.slice_in_dim(x, feature_idx_start, feature_idx_limit, stride, axis=1).astype(dtype)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Automatic modelling routines used by twinify main script.
"""

from collections import OrderedDict

import jax.numpy as jnp
import jax
from jax.dtypes import canonicalize_dtype, issubdtype

import numpyro
import numpyro.distributions as dists
from numpyro.primitives import sample, param, deterministic, plate
from numpyro.handlers import seed, trace
from d3p.util import unvectorize_shape_2d

from .mixture_model import MixtureModel
from .na_model import NAModel

import numpy as np

from typing import Type, Dict, Optional, Union, Tuple, List, Callable, Sequence
from abc import ABCMeta, abstractmethod

import pandas as pd


constraint_dtype_lookup = {
    dists.constraints._Boolean: 'int',
    dists.constraints._Real: 'float',
    dists.constraints._GreaterThan: 'float',
    dists.constraints._Interval: 'float',
    dists.constraints._IntegerGreaterThan: 'int',
    dists.constraints._IntegerInterval: 'int',
    dists.constraints._Multinomial: 'int',
}

class Distribution:
    """ Wraps a numpyro distribution and exposes relevant properties for
        automated model building
    """

    def __init__(self, name: str, numpyro_class: Type[dists.Distribution]) -> None:
        self._name = name
        self._numpyro_class = numpyro_class
        self._is_categorical = numpyro_class in categorical_dist_lookup
        assert (not self.is_categorical or self.is_discrete), \
                "A continuos distribution cannot be categorical"

    @staticmethod
    def from_distribution_instance(dist: dists.Distribution) -> 'Distribution':
        dist_class = type(dist)
        assert(issubclass(dist_class, dists.Distribution))

        try:
            name = dist_class.__name__
        except:
            name = ''

        return Distribution(name, dist_class)

    @property
    def name(self) -> str:
        return self._name

    @property
    def numpyro_class(self) -> Type[dists.Distribution]:
        return self._numpyro_class

    @property
    def parameter_transforms(self) -> Dict[str, dists.transforms.Transform]:
        param_transforms = {
            param: dists.transforms.biject_to(constraint)
            for param, constraint in self._numpyro_class.__dict__['arg_constraints'].items()
        }
        return param_transforms

    def instantiate(self, parameters: Optional[Dict[str, jnp.array]] = None) -> dists.Distribution:

        if parameters is None:
            parameters = self._make_zero_parameters()
        return self._numpyro_class(**parameters)

    def __call__(self, **parameters: Dict[str, jnp.array]) -> dists.Distribution:
        return self.instantiate(parameters)

    def _make_zero_parameters(self):
        zero_params = {
            param: transform(jnp.zeros(1))
            for param, transform in self.parameter_transforms.items()
        }
        return zero_params

    @staticmethod
    def get_support_constraint(dist: Union[Type[dists.Distribution], dists.Distribution]) -> str:
        support_constraint = type(dists.constraints.real)
        if dist.support is not None:
            if not isinstance(dist.support, dists.constraints.Constraint):
                raise ValueError("Support of distribution of type {} cannot \
                    be derived from non-instantiated class".format(dist))
            support_constraint = type(dist.support)
        return support_constraint


    @staticmethod
    def get_support_dtype(dist: Union[Type[dists.Distribution], dists.Distribution]) -> str:
        """ Determines the type of a distribution's support

        Parameters:
            dist (dists.Distribution): numpyro Distribution instance
        Returns:
            (str) dtype of the given distribution's support
        """
        support_constraint = Distribution.get_support_constraint(dist)

        try:
            return canonicalize_dtype(constraint_dtype_lookup[support_constraint])
        except KeyError:
            return ValueError("A distribution with support {} is currently \
                not supported".format(support_constraint))

    @property
    def support_dtype(self) -> str:
        if self.has_data_dependent_shape:
            return self.get_support_dtype(self.instantiate())
        else:
            return self.get_support_dtype(self.numpyro_class)

    @property
    def has_data_dependent_shape(self) -> bool:
        try:
            return dists.constraints.is_dependent(self.numpyro_class.support)
        except AttributeError:
            return True

    @property
    def is_categorical(self) -> bool:
        return self._is_categorical

    @property
    def is_discrete(self) -> bool:
        return self.get_support_constraint(self.numpyro_class).is_discrete


class ModelFeature:
    """ Represents how a feature in the data is modelled
    """

    def __init__(self, feature_name: str, dist: Distribution) -> None:
        self._name = feature_name
        self._dist = dist
        self._shape = None
        self._value_map = None
        self._missing_values = False

    @staticmethod
    def from_distribution_instance(feature_name: str, dist: dists.Distribution) -> 'ModelFeature':
        meta_dist = Distribution.from_distribution_instance(dist)
        shape = dist.batch_shape
        feature = ModelFeature(feature_name, meta_dist)
        feature.shape = shape
        return feature

    @property
    def name(self) -> str:
        """ Feature name """
        return self._name

    @property
    def distribution(self) -> Distribution:
        """ automodel.Distribution the feature is modelled by """
        return self._dist

    @property
    def shape(self) -> Tuple[int]:
        """ Simple shape of parameters for this feature distribution.

        Does not include mixture dimension.
        """
        if not self.distribution.has_data_dependent_shape:
            return tuple()
        elif self._shape is None:
            raise RuntimeError("The shape of the distribution depends on the data\
                but not data has been preprocessed yet.")
        return self._shape

    @shape.setter
    def shape(self, value: Tuple[int]):
        self._shape = value

    @property
    def has_missing_values(self) -> bool:
        return self._missing_values

    def preprocess_data(self, data: Union[pd.DataFrame, pd.Series]) -> Union[pd.DataFrame, pd.Series]:
        """ Preprocesses data according to the feature distribution.

        If a pd.Series instance is passed in, it will be treated as the feature
        data. If a pd.DataFrame is passed in, this function will only affect
        the column corresponding the feature's name (self.name).

        If the feature follows a Categorical distribution, this will extract
        the number of categories present in the data.

        If the data type is not numeric for any distribution modelling some
        kind of categories (e.g., Categorical, Bernoulli), it will be mapped
        to integer categories.
        """
        column = data
        if isinstance(data, pd.DataFrame):
            column = data[self.name]
        assert(isinstance(column, pd.Series))

        # detect whether there are missing values in the data
        if np.any(column.isna()):
            self._missing_values = True

        # TODO: with the introduction of DataDescription, this is probably somewhat redundant now - need to test
        # if the distribution is a categorical, we need to determine the number
        # of categories
        if self.distribution.is_categorical:

            # if shape of the categorical distribution depends on data, we extract the
            #    number of unique feature values in the data
            if self.distribution.has_data_dependent_shape:
                self.shape = (len(column.dropna().unique()), )

            # we map the feature values to integers
            self._value_map = {val: i for i, val in enumerate(column.dropna().unique())}
            column = column.map(self._value_map)

        if isinstance(data, pd.DataFrame):
            data[self.name] = column
        else:
            data = column
        return data

    def postprocess_data(self, data: Union[pd.DataFrame, pd.Series]) -> Union[pd.DataFrame, pd.Series]:
        column = data
        if isinstance(data, pd.DataFrame):
            column = data[self.name]
        assert(isinstance(column, pd.Series))

        if self._value_map is not None:
            inverse_map = {v: k for k, v in self._value_map.items()}
            column = column.map(inverse_map)

        if isinstance(data, pd.DataFrame):
            data[self.name] = column
        else:
            data = column
        return data

    def instantiate(self, **params):
        return self.distribution(**params)

    def __repr__(self) -> str:
        return f"ModelFeature({self.name}, {self.distribution}, {self.shape})"

################################# model parsing ################################


categorical_dist_lookup = {
    dists.BernoulliProbs, dists.BernoulliLogits,
    dists.CategoricalProbs, dists.CategoricalLogits,
    dists.MultinomialProbs, dists.MultinomialLogits,
    dists.OrderedLogistic
}

dist_lookup = {
    "normal": Distribution("Normal", dists.Normal),
    "bernoulli": Distribution("Bernoulli", dists.BernoulliProbs),
    "categorical": Distribution("Categorical", dists.CategoricalProbs),
    "poisson": Distribution("Poisson", dists.Poisson)
}


def make_distribution(token: str) -> Distribution:
    try:
        return dist_lookup[token.lower()]
    except KeyError:
        raise ValueError("{} distributions are currently not supported".format(token))

class ParsingError(Exception):
    pass

def parse_model(model_str: str) -> List[ModelFeature]:
    """
    Parameters:
        model_str (str)
    Returns:
        OrderedDict(feature_name -> Distribution)
    """
    features = []
    for line_nr, line in enumerate(model_str.splitlines()):
        # ignore comments (indicated by #)
        line = line.split('#')[0].strip()
        if len(line) == 0:
            continue

        parts = line.split(':')
        if len(parts) == 2:
            feature_name = parts[0].strip()
            distribution_name = parts[1].strip()

            try:
                dist = make_distribution(distribution_name)
                features.append(ModelFeature(feature_name, dist))
            except ValueError as e:
                raise ParsingError("Unable to parse line {}:\n<< {} >>\n{}".format(
                    line_nr, line, e
                ))

        else:
            raise ParsingError("Unable to parse line {}:\n<< {} >>\nSyntax error!".format(
                line_nr, line
            ))
    return features

def extract_features_from_mixture_model(mixture: MixtureModel, feature_names: List[str]) -> List[ModelFeature]:
    features = [ModelFeature.from_distribution_instance(name, dist)
                for name, dist in zip(feature_names, mixture.distributions)]
    return features


def parse_model_fn(model: Callable, feature_names: List[str]) -> List[ModelFeature]:
    model_trace = trace(seed(model, rng_seed=0)).get_trace(1)
    assert('x' in model_trace)
    mixture_model = model_trace['x']['fn']
    return extract_features_from_mixture_model(mixture_model, feature_names)


##################### prior lookup ########################

prior_lookup = {
    dists.Normal: {'loc': (dists.Normal, (0., 10.)), 'scale': (dists.LogNormal, (0., 2.))},
    dists.BernoulliProbs: {'probs': (dists.Beta, (1., 1.))},
    dists.CategoricalProbs: {'probs': (dists.Dirichlet, (1.,))},
    dists.Poisson: {'rate': (dists.Exponential, (1.,))}
}

def create_feature_prior_dists(feature: ModelFeature, k: int) -> Dict[str, dists.Distribution]:
    shape = (k,) + feature.shape
    prior_dists = {
        parameter: prior_dist(*[jnp.ones(shape) * prior_param for prior_param in prior_params])
        for parameter, (prior_dist, prior_params) in prior_lookup[feature.distribution.numpyro_class].items()
    }
    return prior_dists

################### typed dist ##########################
class TypedDistribution(dists.Distribution):

    def __init__(self, base_dist, dtype, validate_args=None):
        super(TypedDistribution, self).__init__(base_dist.batch_shape, base_dist.event_shape, validate_args=validate_args)
        self.dtype = dtype
        self.base_dist = base_dist

    @property
    def support(self):
        return self.base_dist.support

    def log_prob(self, value):
        log_prob = self.base_dist.log_prob(value.astype(self.dtype))
        return log_prob

    def sample(self, key, sample_shape=()):
        return self.base_dist.sample(key, sample_shape=sample_shape)

    def __repr__(self) -> str:
        return f"TypedDistribution({self.base_dist})"


################### automatically build mixture model ##########################

def make_model(features: List[ModelFeature], k: int) -> Callable[..., None]:
    """ Given model feature specifications, create a model function for a mixture model

    Parameters:
        features: list of ModelFeature instances, specifying the feature distributions
        k: number of mixture components
    """
    def model(x=None, num_obs_total=None):
        assert x is None or len(jnp.shape(x)) == 2
        if x is None:
            N = 1
        else:
            N = jnp.shape(x)[0]
        if num_obs_total is None:
            num_obs_total = N

        assert isinstance(num_obs_total, int) and num_obs_total > 0
        assert N <= num_obs_total

        mixture_dists = []
        for feature in features:
            prior_values = {}
            feature_prior_dists = create_feature_prior_dists(feature, k)
            for feature_prior_param, feature_prior_dist in feature_prior_dists.items():
                prior_values[feature_prior_param] = sample(
                    "{}_{}".format(feature.name, feature_prior_param),
                    feature_prior_dist
                )

            feature_dist = feature.instantiate(**prior_values)
            feature_dist = TypedDistribution(feature_dist, feature.distribution.support_dtype)
            if feature.has_missing_values:
                feature_na_prob = sample(
                    "{}_na_prob".format(feature.name),
                    dists.Beta(2.*jnp.ones(k), 2.*jnp.ones(k))
                )
                feature_dist = NAModel(feature_dist, feature_na_prob)

            mixture_dists.append(feature_dist)

        pis = sample('pis', dists.Dirichlet(jnp.ones(k)))
        with plate('batch', num_obs_total, N):
            mixture_model_dist = MixtureModel(mixture_dists, pis)
            x = sample('x', mixture_model_dist, obs=x)
            return x
    return model

####################### postprocessing ###########################
def postprocess_function_factory(features):
    def postprocess_fn(syn_df: pd.DataFrame) -> pd.DataFrame:
        encoded_syn_df = syn_df.copy()
        for feature in features:
            encoded_syn_df = feature.postprocess_data(encoded_syn_df)
        return encoded_syn_df
    return postprocess_fn



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable
ModelFunction = Callable
GuideFunction = Callable

from collections import namedtuple
PrivacyLevel = namedtuple("PrivacyLevel", ["epsilon", "delta", "dp_noise"])


from twinify.dpvi.dpvi_model import DPVIModel, InferenceException
from twinify.dpvi.dpvi_result import DPVIResult

import twinify.dpvi.modelling



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Visualization routines used by twinify main script.
"""

import numpy as onp
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from mpl_toolkits.axes_grid1 import ImageGrid
from collections import OrderedDict as od

syn_data_color = 'red'
syn_data_label = 'Synthetic data'
orig_data_color = 'blue'
orig_data_label = 'Original data'

def plot_margins(syn_df, orig_df, show=False):
    fig, axis = plt.subplots()
    red_patch = mpatches.Patch(color=syn_data_color, label=syn_data_label)
    blue_patch = mpatches.Patch(color=orig_data_color, label=orig_data_label)
    axis.legend(handles=[blue_patch, red_patch])

    for location, name in enumerate(syn_df.columns):
        col = syn_df[name].dropna().values
        if col.size == 0:
            continue
        syn_violin_part = axis.violinplot(col, [location])
        syn_violin_part['bodies'][0].set_color(syn_data_color)
        syn_violin_part["cmaxes"].set_color(syn_data_color)
        syn_violin_part["cmins"].set_color(syn_data_color)
        syn_violin_part["cbars"].set_color(syn_data_color)

    orig_violin_parts = od()
    for location, name in enumerate(syn_df.columns):
        orig_violin_parts[name] = axis.violinplot(orig_df[name].dropna().values, [location])
    for orig_violin_part in orig_violin_parts.values():
        body = orig_violin_part["bodies"][0]
        body.set_color(orig_data_color)
        orig_violin_part["cmaxes"].set_color(orig_data_color)
        orig_violin_part["cmins"].set_color(orig_data_color)
        orig_violin_part["cbars"].set_color(orig_data_color)

    axis.set_xticks(onp.arange(0, syn_df.shape[1]))
    axis.set_xticklabels(list(syn_df.columns), rotation=45)
    if show:
        fig.show()
    return fig

def plot_covariance_heatmap(syn_df, orig_df, show=False):
    syn_cov = syn_df.cov().values
    orig_cov = orig_df.cov().values
    cov_max = onp.maximum(onp.max(orig_cov[onp.isfinite(orig_cov)]), onp.max(syn_cov[onp.isfinite(syn_cov)]))
    cov_min = onp.minimum(onp.min(orig_cov[onp.isfinite(orig_cov)]), onp.min(syn_cov[onp.isfinite(syn_cov)]))

    fig = plt.figure()
    grid = ImageGrid(fig, 111, nrows_ncols=(1,2), axes_pad=0.15, share_all=True, cbar_location="right", cbar_mode="single", cbar_pad=0.15)

    grid[0].imshow(orig_cov, vmin=cov_min, vmax=cov_max)
    grid[0].set_xticks(onp.arange(0, syn_df.shape[1]))
    grid[0].set_yticks([])
    grid[0].set_xticklabels(list(syn_df.columns), rotation=45, fontdict={"fontsize":10})
    grid[0].set_title(orig_data_label)

    im = grid[1].imshow(syn_cov, vmin=cov_min, vmax=cov_max)
    grid[1].set_xticks(onp.arange(0, syn_df.shape[1]))
    grid[1].set_yticks(onp.arange(0, syn_df.shape[1]))
    grid[1].set_xticklabels(list(syn_df.columns), rotation=45, fontdict={"fontsize":10})
    grid[1].set_yticklabels(list(syn_df.columns), fontdict={"fontsize":10})
    grid[1].set_title(syn_data_label)

    grid[1].cax.colorbar(im)
    grid[1].cax.toggle_label(True)

    if show:
        fig.show()
    return fig

def plot_missing_values(syn_df, orig_df, show=False):
    syn_nas = syn_df.isna().mean()
    orig_nas = orig_df.isna().mean()
    bar_width = 0.25
    fig, axis = plt.subplots()
    num_features = syn_df.shape[1]
    x_ticks = onp.arange(num_features)
    axis.bar(x_ticks-.5*bar_width, syn_nas, width=bar_width, color=syn_data_color, edgecolor='white', label=syn_data_label)
    axis.bar(x_ticks+.5*bar_width, orig_nas, width=bar_width, color=orig_data_color, edgecolor='white', label=orig_data_label)
    axis.set_xticks(x_ticks)
    axis.set_xticklabels(list(syn_nas.index), rotation=45)

    red_patch = mpatches.Patch(color=syn_data_color, label=syn_data_label)
    blue_patch = mpatches.Patch(color=orig_data_color, label=orig_data_label)
    axis.legend(handles=[blue_patch, red_patch])

    if show:
        fig.show()
    return fig


# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List, Union, Mapping, Iterable, Any, Dict, TypeVar
from pandas.api.types import is_integer_dtype, is_categorical_dtype, is_float_dtype, is_string_dtype, is_bool_dtype
import pandas as pd
import itertools
from functools import reduce
from operator import mul
import numpy as np
import logging

logger = logging.getLogger(__name__)

Dtype = TypeVar("Dtype")


class DataDescription:

    def __init__(self, dtypes: Mapping[str, Dtype]) -> None:
        self._dtypes = dict(dtypes)

    @staticmethod
    def from_dataframe(df: pd.DataFrame, strings_to_categories: bool = True) -> "DataDescription":
        dtypes = dict()
        for col in df.columns:
            dtype = df[col].dtype
            if is_string_dtype(dtype):
                if not strings_to_categories:
                    raise ValueError(
                        f"Input may not contain columns of dtype string unless strings_to_categories is True. Column: {col}.")

                try:
                    dtype = pd.CategoricalDtype(df[col].dropna().unique())
                except ValueError as e:
                    raise Exception(f"Cannot interpet type of column {col}", e)
            if not (is_categorical_dtype(dtype) or is_integer_dtype(dtype) or is_float_dtype(dtype) or is_bool_dtype(dtype)):
                raise ValueError(f"Only float, integer or categorical dtypes are currently supported, but column {col} has dtype {dtype}.")

            dtypes[col] = dtype

        return DataDescription(dtypes)

    @property
    def columns(self) -> Tuple[str]:
        return tuple(self._dtypes.keys())

    @property
    def num_columns(self) -> int:
        return len(self.columns)

    @property
    def dtypes(self) -> Dict[str, Dtype]:
        return self._dtypes.copy()

    @property
    def all_columns_discrete(self) -> bool:
        return np.all([
            not is_float_dtype(dtype) for dtype in self._dtypes.values()
        ])

    @staticmethod
    def _map_column_to_numeric(x: pd.Series) -> pd.Series:
        assert (
                is_categorical_dtype(x.dtype) or
                is_float_dtype(x.dtype) or
                is_integer_dtype(x.dtype) or
                is_bool_dtype(x.dtype)
        )
        if is_categorical_dtype(x.dtype):
            return x.cat.codes
        return x

    def map_to_numeric(self, categorical_df: pd.DataFrame) -> pd.DataFrame:
        numeric_df = categorical_df.copy()
        for col, dtype in self._dtypes.items():
            if col not in categorical_df.columns:
                continue

            numeric_df[col] = self._map_column_to_numeric(numeric_df[col].astype(dtype))
        return numeric_df

    @staticmethod
    def _are_all_values_integers(x: pd.Series) -> bool:
        return np.allclose(x - x.astype(int), 0.)

    def map_to_categorical(self, numeric_df_or_array: Union[pd.DataFrame, np.ndarray]) -> pd.DataFrame:
        if isinstance(numeric_df_or_array, pd.DataFrame):
            numeric_df = numeric_df_or_array
        else:
            shape = np.shape(numeric_df_or_array)
            if len(shape) != 2 or shape[1] != len(self.columns):
                raise ValueError(
                    f"Array must be two-dimensional with a second dimension of size {len(self.columns)}, had shape {shape}.")
            numeric_df = pd.DataFrame(numeric_df_or_array, columns=self.columns)

        categorical_df = pd.DataFrame()

        for col, dtype in self._dtypes.items():
            if col not in numeric_df.columns:
                continue

            col_values = numeric_df[col]
            if is_categorical_dtype(dtype):
                if not self._are_all_values_integers(col_values):
                    raise ValueError(f"Cannot map column {col} to categories because the input were no integers.")

                categorical_df[col] = pd.Categorical.from_codes(col_values.astype(int), dtype=dtype)
            else:
                categorical_df[col] = col_values.astype(dtype)
        return categorical_df

    def __eq__(self, other) -> bool:
        return isinstance(other, DataDescription) and self._dtypes == other._dtypes


class DataFrameData:
    """Converter between categorical and integer formatted dataframes."""

    def __init__(self, base_df: pd.DataFrame):
        """Initialise.
        Args:
            base_df (DataFrame): Base categorical dataframe.
        """
        self._data_description = DataDescription.from_dataframe(base_df)
        if not self._data_description.all_columns_discrete:
            raise ValueError(
                "Data has columns which are not discrete, i.e., neither of type bool, integer, or categorical.")

        self.int_df = self._data_description.map_to_numeric(base_df)

        self._values_by_col = {
            col: list(range(len(base_df[col].cat.categories))) if is_categorical_dtype(base_df[col])
            else sorted(list(base_df[col].unique()))
            for col in self.int_df.columns
            if (is_categorical_dtype(self.int_df[col]) or is_integer_dtype(self.int_df[col]))
        }
        # self.values_by_int_feature = { #TODO: rename to values_by_feature_idx if used anywhere?
        #     i: list(self._values_by_col[col])
        #     for i, col in enumerate(self.int_df.columns)
        #     if (is_categorical_dtype(self.int_df[col]) or is_integer_dtype(self.int_df[col]))
        # }

    @property
    def values_by_col(self) -> Dict[str, List[int]]:
        return self._values_by_col

    @property
    def columns(self) -> Iterable[str]:
        return self.int_df.columns

    @property
    def data_description(self) -> DataDescription:
        return self._data_description

    def get_x_values(self) -> np.ndarray:
        """Enumerate all possible datapoints for the categories of the dataset.
        Returns:
            np.ndarray: Enumeration of all possible datapoints.
        """
        x_values = np.zeros((self.get_domain_size(), len(self.columns)))
        for i, val in enumerate(itertools.product(*self.values_by_col.values())):
            x_values[i, :] = np.array(val)
        return x_values

    def get_domain_size(self) -> int:
        """Compute the number of possible datapoints in the domain of the dataset.
        Returns:
            int: The number of possible datapoints in the domain.
        """
        return reduce(mul, [len(col_values) for col_values in self.values_by_col.values()])

    def int_df_to_cat_df(self, int_df: pd.DataFrame) -> pd.DataFrame:
        """Convert integer-valued dataframe to categorical dataframe.
        Args:
            int_df (DataFrame): The integer valued dataframe.
        Returns:
            DataFrame: Categorical valued dataframe.
        """
        return self._data_description.map_to_categorical(int_df)

    def ndarray_to_cat_df(self, ndarray: np.ndarray) -> pd.DataFrame:
        """Convert integer-valued ndarray to categorical dataframe.
        Args:
            ndarray (ndarray): The integer-valued array to convert.
        Returns:
            DataFrame: The categorical dataframe.
        """
        return self._data_description.map_to_categorical(ndarray)

    def int_query_to_str_query(self, inds: Iterable[int], value: Iterable[int]) -> Tuple[List[str], List[str]]:
        """Convert marginal query for integer dataframe to query for categorical dataframe.
        Args:
            inds (tuple: Query indices.
            value (tuple or np.ndarray): Query value.
        Returns:
            (List, List): String-valued indices and value.
        """
        value = tuple(value) if isinstance(value, np.ndarray) else value
        column_names = [self._data_description.columns[ind] for ind in inds]
        str_values = [
            self._data_description.dtypes[column_names[i]].categories[v]
            if is_categorical_dtype(self._data_description.dtypes[column_names[i]])
            else v
            for i, v in enumerate(value)
        ]
        return column_names, str_values

    def str_query_to_int_query(self, feature_set: Iterable[str], value: List[str]) -> Tuple[List, np.ndarray]:
        """Convert marginal query for categorical dataframe to query for integer dataframe.
        Args:
            feature_set (tuple): Query indices.
            value (tuple): Query values.
        Returns:
            (tuple, tuple): Converted query indices and value.
        """

        def index(list, value):
            if value in list:
                return list.index(value)
            else:
                raise ValueError("{} not in {}".format(value, list))

        int_inds = [index(self._data_description.columns, feature) for feature in feature_set]
        int_values = [
            pd.Categorical(value[i], dtype=self._data_description.dtypes[feature]).codes
            if is_categorical_dtype(self._data_description.dtypes[feature])
            else value[i]
            for i, feature in enumerate(feature_set)
        ]
        return int_inds, np.array(int_values)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from typing import Optional, Any, Tuple

import numpy as np
import jax
import jax.numpy as jnp
import jax.experimental.host_callback as hcb

import d3p.minibatch
import d3p.random
import d3p.dputil
import d3p.svi
import numpyro.infer
from twinify.base import InferenceModel
from tqdm import tqdm

from twinify.dpvi import PrivacyLevel, ModelFunction, GuideFunction
from twinify.dpvi.dpvi_result import DPVIResult
from twinify.dataframe_data import DataDescription


class InferenceException(Exception):

    def __init__(self, iteration: int, total_iterations: int) -> None:
        self._iteration = iteration
        self._total_iterations = total_iterations
        super().__init__(
            f"Inference encountered NaN value after {iteration}/{total_iterations} iterations {self.progress*100:.3f} %."
        )

    @property
    def iteration(self) -> int:
        return self._iteration

    @property
    def total_iterations(self) -> int:
        return self._total_iterations

    @property
    def progress(self) -> float:
        return self._iteration / self._total_iterations


class SilenceableProgressBar:

    def __init__(self, total: int, silent: bool) -> None:
        self._silent = silent
        if not self._silent:
            self._tqdm = tqdm(total=total, desc="epochs")

    @staticmethod
    def static_update(tqdm, silent: bool, loss: float) -> None:
        if not silent:
            tqdm.set_postfix_str(f"ELBO {loss:.3f}")
            tqdm.update(1)

    def update(self, loss: float) -> None:
        self.static_update(self._tqdm, self._silent, loss)

    def update_from_jax(self, loss: float) -> None:
        def _update_callback(loss, transforms):
            self.static_update(self._tqdm, self._silent, loss)

        if not self._silent:
            hcb.id_tap(_update_callback, loss)

    def close(self) -> None:
        if not self._silent:
            self._tqdm.close()


class DPVIModel(InferenceModel):

    def __init__(
            self,
            model: ModelFunction,
            guide: Optional[GuideFunction] = None,
            clipping_threshold: float = 1.,
            num_epochs: int = 1000, # TODO: ultimately this should be num_iters for more flexibility, but the whole pipeline is currently geared towards num_epochs
            subsample_ratio: float = 0.01,
        ) -> None:
        """
        Initialises a probabilistic model for performing differentially-private
        variational inference.

        Args:
            model (ModelFunction): A numpyro model function that programmatically describes the probabilistic model
                which generates the data.
            guide (GuideFunction): Optional numpyro function that programmatically describes the variational approximation
                to the true posterior distribution.
            clipping_threshold (float): The maximum allowed norm of per-example gradients.
            num_epochs (int): How many passes to make over the data.
            subsample_ratio (float): How large of a fraction of the data will constitute a single update batch.
        """
        super().__init__()
        self._model = model

        if guide is None:
            guide = self.create_default_guide(model)

        self._guide = guide
        self._clipping_threshold = clipping_threshold
        self._num_epochs = num_epochs
        self._subsample_ratio = subsample_ratio

    @staticmethod
    def create_default_guide(model: ModelFunction) -> GuideFunction:
        return numpyro.infer.autoguide.AutoDiagonalNormal(model)

    def fit(self,
            data: pd.DataFrame,
            rng: d3p.random.PRNGState,
            epsilon: float,
            delta: float,
            silent: bool = False,
            verbose: bool = False) -> DPVIResult:

        q = self._subsample_ratio
        num_epochs = self._num_epochs

        num_data = data.shape[0]
        batch_size = np.maximum(self.batch_size_for_subsample_ratio(q, num_data), 1)
        q = np.maximum(self.subsample_ratio_for_batch_size(batch_size, num_data), q)
        num_iter = self.num_iterations_for_epochs(num_epochs, q)
        dp_scale, act_epsilon, _ = d3p.dputil.approximate_sigma(epsilon, delta, q, num_iter, maxeval=20)

        if verbose:
            scale_per_sample = dp_scale / batch_size
            print(f"Applying noise with std deviation {dp_scale:.2f} (~ {scale_per_sample:.2f} per element in batch)"
                f" to achieve privacy epsilon of {act_epsilon:.3f} (for delta {delta:.2e}). "
            )
            # TODO: warn for high noise? but when is it too high? what is a good heuristic?


        optimizer = numpyro.optim.Adam(1e-3)

        svi = d3p.svi.DPSVI(
            self._model, self._guide,
            optimizer, numpyro.infer.Trace_ELBO(),
            num_obs_total=num_data, clipping_threshold=self._clipping_threshold,
            dp_scale=dp_scale, rng_suite=d3p.random
        )

        svi_rng, init_batch_rng, epochs_rng = d3p.random.split(rng, 3)

        data_description = DataDescription.from_dataframe(data)

        data = np.asarray(data_description.map_to_numeric(data))
        init_batching, get_batch = d3p.minibatch.subsample_batchify_data((data,), batch_size, rng_suite=d3p.random)
        _, batchify_state = init_batching(init_batch_rng)

        batch = get_batch(0, batchify_state)
        svi_state = svi.init(svi_rng, *batch)

        bar = SilenceableProgressBar(num_epochs, silent)

        def is_nan_in_state(svi_state) -> bool:
            params = svi.get_params(svi_state)
            return jnp.logical_not(
                jnp.all(jnp.array(
                    jax.tree_util.tree_leaves(
                        jax.tree_util.tree_map(
                            lambda x: jnp.logical_not(jnp.any(jnp.isnan(x))),
                            params
                        )
                    )
                ))
            )

        @jax.jit
        def loop_cond(args):
            """ While loop condition. Stop if:
                - have reached desired number of epochs
                - any inferred parameter is NaN, indicating failure -> abort
            """
            cur_epoch, svi_state, _ = args
            no_nans = jnp.logical_not(is_nan_in_state(svi_state))
            return jnp.logical_and(cur_epoch < num_epochs, no_nans)

        @jax.jit
        def train_epoch(args):
            e, svi_state, _ = args

            batchify_rng = d3p.random.fold_in(epochs_rng, e)
            num_batches, batchify_state = init_batching(batchify_rng)

            def train_iteration(i, state_and_loss):
                svi_state, loss = state_and_loss
                batch = get_batch(i, batchify_state)
                svi_state, iter_loss = svi.update(svi_state, *batch)
                return (svi_state, loss + iter_loss / (num_batches * num_data))

            new_svi_state, new_loss = jax.lax.fori_loop(0, num_batches, train_iteration, (svi_state, 0.))

            bar.update_from_jax(new_loss)

            return e + 1, new_svi_state, new_loss

        final_epoch, svi_state, loss = jax.lax.while_loop(loop_cond, train_epoch, (0, svi_state, 0.))

        if is_nan_in_state(svi_state):
            raise InferenceException(final_epoch + 1, num_epochs)

        bar.close()

        params = svi.get_params(svi_state)
        return DPVIResult(
            self._model, self._guide,
            params,
            PrivacyLevel(epsilon, delta, dp_scale),
            loss,
            data_description
        )

    @staticmethod
    def num_iterations_for_epochs(num_epochs: int, subsample_ratio: float) -> int:
        return int(num_epochs / subsample_ratio)

    @staticmethod
    def num_epochs_for_iterations(num_iterations: int, subsample_ratio: float) -> int:
        return int(np.ceil(num_iterations * subsample_ratio))

    @staticmethod
    def batch_size_for_subsample_ratio(subsample_ratio: float, num_data: int) -> int:
        return d3p.minibatch.q_to_batch_size(subsample_ratio, num_data)

    @staticmethod
    def subsample_ratio_for_batch_size(batch_size: int, num_data: int) -> float:
        return d3p.minibatch.batch_size_to_q(batch_size, num_data)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Optional

import numpy as np
import scipy.stats as stats


def compute_aggregates(q: np.ndarray, u: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute aggregates for Rubin's rules from point and variance estimates.
    Args:
        q (ndarray): Point estimates.
        u (ndarray): Variance estimates.
    Returns:
        tuple: (q_m, b_m, u_m).
    """
    q_m = np.mean(q)
    b_m = np.var(q, ddof=1)
    u_m = np.mean(u)

    return q_m, b_m, u_m


def conf_int(q: np.ndarray, u: np.ndarray, conf_level: float) -> np.ndarray:
    """Compute confidence interval with Rubin's rules.
    Args:
        q (ndarray): Point estimates.
        u (ndarray): Variance estimates.
        conf_level (float): Confidence level.
    Returns:
        ndarray: The confidence interval.
    """
    dist = conf_int_distribution(q, u)
    return dist.interval(conf_level) if dist is not None else np.repeat(np.nan, 2)


def non_negative_conf_int(q: np.ndarray, u: np.ndarray, conf_level: float, n: int, n_orig: int) -> np.ndarray:
    """Compute confidence interval with Rubin's rules and non-negative variance estimate.
    Args:
        q (ndarray): Point estimates.
        u (ndarray): Variance estimates.
        conf_level (float): Confidence level.
        n (int): Number of synthetic datapoints.
        n_orig (int): Number of original datapoints.
    Returns:
        ndarray: The confidence interval.
    """
    dist = conf_int_distribution(q, u, True, n, n_orig)
    return dist.interval(conf_level) if dist is not None else np.repeat(np.nan, 2)


def conf_int_distribution(q: np.ndarray, u: np.ndarray, use_nonnegative_variance: Optional[bool] = False,
                          n: Optional[int] = None, n_orig: Optional[int] = None) -> Optional[stats.t]:
    """Compute the estimator distribution with Rubin's rules used for confidence intervals and hypothesis tests.
    Args:
        q (ndarray, 1d): Point estimates.
        u (ndarray, 1d): Variance estimates.
        use_nonnegative_variance (bool, optional): Use the non-negative variance estimate. Defaults to False.
        n (int, optional): Number of synthetic datapoints. Required with non-negative variance estimate. Defaults to None.
        n_orig (int, optional): Number of original datapoints. Required with non-negative variance estimate. Defaults to None.
    Returns:
        scipy distribution: The distribution as a scipy.stats distribution object.
    """
    q_m, b_m, u_m = compute_aggregates(q, u)
    m = q.size
    T_m = (1 + 1 / m) * b_m - u_m
    if use_nonnegative_variance:
        T_m = T_m if T_m > 0 else n / n_orig * u_m
    degree = (m - 1) * (1 - u_m / ((1 + 1 / m) * b_m)) ** 2
    if not np.isfinite(degree):
        print("m: {}, u_m: {}, b_m: {}".format(m, u_m, b_m))

    if T_m < 0:
        return None
    else:
        return stats.t(loc=q_m, scale=np.sqrt(T_m), df=degree)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import graphviz
from typing import Tuple, List, Optional, Set, Dict, Union, TypeVar, Iterable

T = TypeVar('T')


class TreeNode:
    def __init__(self, variables: 'TreeNode', parent: Optional['TreeNode']):
        self.variables = variables
        self.parent = parent
        self.children = []
        self.reset()

    def reset(self):
        self.potential = None
        self.messages = []
        self.result = None


class EmptyFactor:
    def __init__(self, scope: Union[Tuple[T, ...], Set[T]]):
        self.scope: Set[T] = set(scope)
        self.origin = None

    def product(self, other: 'EmptyFactor') -> 'EmptyFactor':
        return EmptyFactor(self.scope.union(other.scope))

    def marginalise(self, variable) -> 'EmptyFactor':
        return EmptyFactor(self.scope.difference({variable}))

    @staticmethod
    def list_product(factors: Iterable['EmptyFactor']) -> 'EmptyFactor':
        factors = list(factors)
        product = factors[0]
        for i in range(1, len(factors)):
            product = product.product(factors[i])
        return product


class JunctionTree:
    """Implementation of a junction tree.
    This method does not check that the given nodes and edges actually define
    a valid junction tree.
    """

    def __init__(self, nodes: Set['TreeNode'], edges: Dict[Tuple['TreeNode', 'TreeNode'], Set['TreeNode']],
                 cliques: Iterable[Tuple]):
        """Create the junction tree explicitly.
        Args:
            nodes (set): The nodes of the junction tree.
            edges (dict): Edges as a dict with pair of nodes as keys and sets of nodes (separators) as values.
            cliques (list(tuple)): List of the factor scopes of the underlying Markov network.
        """
        self.nodes = nodes
        self.edges = edges
        self.cliques = cliques

    # Only used with belief propagation, consider removing this with belief propagation
    def init_factor_assignments(self) -> None:
        factor_assignment = {}
        for scope in self.cliques:
            for node in self.nodes:
                if set(scope).issubset(node):
                    if node not in factor_assignment.keys():
                        factor_assignment[node] = []
                    factor_assignment[node].append(scope)
                    break
        self.factors_in_node = factor_assignment
        self.node_for_factor = {}
        for node, factors in self.factors_in_node.items():
            for factor in factors:
                self.node_for_factor[factor] = node

    # Only used with belief propagation, consider removing this with belief propagation
    def init_node_orders(self) -> None:
        self.root_variables: 'TreeNode' = list(self.nodes)[0]
        self.downward_order = self.compute_downward_order(self.root_variables)
        self.root_node = self.downward_order[0]
        self.upward_order = self.downward_order[::-1]

    def check_nodes_edges(self) -> None:
        for e1, e2 in self.edges.keys():
            edge = (e1, e2)
            if e1 not in self.nodes:
                raise ValueError("{} from {} not in nodes".format(e1, edge))
            if e2 not in self.nodes:
                raise ValueError("{} from {} not in nodes".format(e2, edge))
            if (e2, e1) in self.edges:
                raise ValueError("Duplicate edge {}".format(edge))
            if self.edges[edge] != set(e1).intersection(set(e2)):
                raise ValueError(
                    "Separator {} is not the intersection of {} and {}".format(set(e1).intersection(set(e2)), e1, e2))

    def check_acyclicity(self) -> None:
        marked_nodes = set()
        node_stack = [list(self.nodes)[0]]
        while len(node_stack) > 0:
            node = node_stack.pop()
            if node in marked_nodes:
                raise ValueError("Detected cycle at node {}".format(node))
            marked_nodes.add(node)
            neighbours = self.get_neighbours(node)
            for neighbour in neighbours:
                if neighbour not in marked_nodes:
                    node_stack.append(neighbour)

    @staticmethod
    def from_variable_elimination(feature_sets: Iterable[Tuple[T, T]], elimination_order: Iterable) -> 'JunctionTree':
        """Create a junction tree from a variable elimination run.
        Args:
            feature_sets (list(tuple)): The factor scopes of the variable elimination.
            elimination_order (list): The elimination order.
        Returns:
            JunctionTree: The resulting JunctionTree.
        """
        factors = [EmptyFactor(feature_set) for feature_set in feature_sets]
        nodes = set()
        edges = {}

        for variable in elimination_order:
            factors = JunctionTree.eliminate_var(factors, variable, nodes, edges)

        jt = JunctionTree(nodes, edges, feature_sets)
        return jt

    @staticmethod
    def eliminate_var(factors: Iterable, variable: Iterable, nodes, edges) -> List:
        factors_in_prod = [factor for factor in factors if variable in factor.scope]
        factors_not_in_prod = [factor for factor in factors if variable not in factor.scope]

        prod_factor = EmptyFactor.list_product(factors_in_prod)
        prod_scope = tuple(prod_factor.scope)
        nodes.add(prod_scope)

        for factor in factors_in_prod:
            if factor.origin is not None:
                edges[(factor.origin, prod_scope)] = set(prod_scope).intersection(set(factor.origin))

        summed_factor = prod_factor.marginalise(variable)
        summed_factor.origin = prod_scope
        factors_not_in_prod.append(summed_factor)
        return factors_not_in_prod

    def visualize(self) -> graphviz.Graph:
        """Visualize the junction tree with Graphviz.
        Returns:
            Graphviz graph: The resulting Graphviz object.
        """
        g = graphviz.Graph()
        for (e1, e2), sep in self.edges.items():
            g.edge(str(e1), str(e2), label=str(sep))
        return g

    def get_neighbours(self, node: 'TreeNode') -> List['TreeNode']:
        return [other for other in self.nodes if (node, other) in self.edges or (other, node) in self.edges]

    def add_edge(self, edge) -> None:
        e1, e2 = edge
        self.edges[edge] = set(e1).intersection(set(e2))

    def remove_redundant_nodes(self) -> None:
        """Prune the juction tree to remove redundant nodes."""
        while True:
            res = self.find_redundant_node()
            if res is not None:
                node, large_neighbour, neighbours = res
                self.nodes.remove(node)
                self.edges = {(e1, e2): sep for (e1, e2), sep in self.edges.items() if e1 != node and e2 != node}
                for neighbour in neighbours:
                    if neighbour != large_neighbour:
                        self.add_edge((large_neighbour, neighbour))

            else:
                self.check_nodes_edges()
                self.init_factor_assignments()
                self.init_node_orders()
                return

    def find_redundant_node(self) -> Tuple['TreeNode', 'TreeNode', List['TreeNode']]:
        for node in self.nodes:
            neighbours = self.get_neighbours(node)
            for neighbour in neighbours:
                if set(node).issubset(set(neighbour)):
                    return node, neighbour, neighbours

    def compute_downward_order(self, root_node: 'TreeNode') -> List:
        order = []
        marked_nodes = set()
        root = TreeNode(root_node, None)
        node_stack = [root]
        while len(node_stack) > 0:
            node = node_stack.pop()
            order.append(node)
            marked_nodes.add(node.variables)
            for neighbour in self.get_neighbours(node.variables):
                if neighbour not in marked_nodes:
                    new_node = TreeNode(neighbour, node)
                    node_stack.append(new_node)
                    node.children.append(new_node)

        return order



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, Tuple, Any
import warnings
import numpy as np
from scipy import special
import scipy.optimize as optim


def delta(epsilon: float, sens_per_sigma: float) -> float:
    """Compute delta for given epsilon and sensitivity per noise standard deviation for the Gaussian mechanism.
    Args:
        epsilon (float)
        sens_per_sigma (float): Sensitivity per noise standard deviation.
    Returns:
        float: Delta
    """
    if sens_per_sigma < 0:
        warnings.warn(
            f"Sensitivity per sigma was negative: {sens_per_sigma}. "
            f"Sensitivity per sigma should be non-negative. Defaulting to 0"
        )

    if sens_per_sigma <= 0:
        return 0
    mu = sens_per_sigma ** 2 / 2
    term1 = special.erfc((epsilon - mu) / np.sqrt(mu) / 2)
    term2 = np.exp(epsilon) * special.erfc((epsilon + mu) / np.sqrt(mu) / 2)
    return 0.5 * (term1 - term2)


def find_sens_per_sigma(epsilon: float, delta_bound: float, sens_per_sigma_upper_bound: Optional[float] = 20) -> Union[
    float, Tuple[float, Any]]:
    """Find the required sensitivity per noise standard deviation for (epsilon, delta)-DP with Gaussian mechanism.
    Args:
        epsilon (float)
        delta_bound (float)
        sens_per_sigma_upper_bound (float, optional): Upper bound guess on sensitivity per sigma. Defaults to 20.
    Returns:
        float: The required sensitivity per noise standard deviation.
    """
    return optim.brentq(lambda sigma: delta(epsilon, sigma) - delta_bound, 0, sens_per_sigma_upper_bound)


def sigma(epsilon: float, delta_bound: float, sensitivity: float, sigma_upper_bound: Optional[float] = 20) -> float:
    """Find the required noise standard deviation for the Gaussian mechanism with (epsilon, delta)-DP.
    Args:
        epsilon (float)
        delta_bound (float)
        sensitivity (float): Sensitivity of the Gaussian mechanism.
        sigma_upper_bound (float, optional): Guess for an upper bound on sensitivity / sigma. Defaults to 20.
    Returns:
        float: The required noise standard deviation.
    """
    return sensitivity / find_sens_per_sigma(epsilon, delta_bound, sigma_upper_bound)


def eps_delta_budget_to_rho_budget(epsilon: float, delta: float) -> float:
    """Convert (epsilon, delta)-DP to rho-zCDP.
    Args:
        epsilon (float)
        delta (float)
    Returns:
        float: rho
    """
    beta = np.sqrt(-np.log(delta))
    return (np.sqrt(beta ** 2 + epsilon) - beta) ** 2
